# -*- coding: utf-8 -*-
import numpy as np
from setuptools import Extension, setup

setup(
  ext_modules=[
    Extension(
     'antaress.Gauss_star_grid',
      sources=['src/antaress/ANTARESS_analysis/C_grid/Gauss_star_grid.c'],
      language='c',
    )
  ],
  include_dirs=[np.get_include()],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from copy import deepcopy
import numpy as np
import pandas as pd
import bindensity as bind
from lmfit import Parameters
from scipy.fft import fft
from scipy.fft import fft2
from ..ANTARESS_general.minim_routines import call_lmfit
from ..ANTARESS_grids.ANTARESS_coord import excl_plrange
from ..ANTARESS_process.ANTARESS_data_align import align_data
from ..ANTARESS_general.utils import dataload_npz,stop,closest,gen_specdopshift,np_where1D,check_data
from ..ANTARESS_general.constant_data import c_light



def detrend_prof(detrend_prof_dic,data_dic,coord_dic,inst,vis,CCF_dic,data_prop,gen_dic,plot_dic):
    r"""**Detrending: main routine.**
    
    Corrects disk-integrated line profiles for abnormal variations
    
     - variations in line contrast, FWHM, and RV should first be fitted within the `plot_dic['prop_DI']` routine to derive their model.
       Model coefficients can then be set in the detrending module.
     - better results are obtained by correcting first for the constrast, before the RV.
       FWHM must be corrected after RVs, so that line profiles are aligned
     - the detrending module is called before the analysis module so that corrected profiles can then be re-analyzed
     - the detrending module requires that aligned profiles have already been calculated, as the RV model has been fitted to profiles corrected for the star motion.
       After correction the routine resets the path of current (and aligned) disk-integrated data to corrected profiles, which will not go through the alignment module again 
     - for modulations the coefficient of degree 0 is not required, as models are normalized to the mean

    Args:
        TBD
    
    Returns:
        TBD
        
    """
    cond_trend = (detrend_prof_dic['corr_trend'] and (inst in detrend_prof_dic['prop']) and (vis in detrend_prof_dic['prop'][inst]))
    cond_PC = (detrend_prof_dic['corr_PC'] and (inst in detrend_prof_dic['PC_model']) and (vis in detrend_prof_dic['PC_model'][inst]))

    #Specific instrumental correction
    if (inst=='EXPRES') and (gen_dic['star_name']=='55Cnc'):cond_custom = True
    else:cond_custom = False

    if cond_trend or cond_PC or cond_custom:
        data_vis=data_dic[inst][vis]
        print('   > Correcting disk-integrated line profiles')
        
        #Calculating aligned data
        if gen_dic['calc_detrend_prof']:
            print('         Calculating data')
            data_prop_vis=data_prop[inst][vis]
            pl_loc = data_vis['transit_pl'][0]
            
            #Correct spectrum
            if ('spec' in data_dic[inst][vis]['type']):
                
                #Single line, processed in RV space
                if detrend_prof_dic['line_trans'] is not None:
                    single_l = True
                    iord_line = detrend_prof_dic['iord_line']
                    nord_eff = 1
                    dim_exp_eff=[nord_eff,data_dic[inst][vis]['nspec']]  
                    
                #Correct full spectrum
                else:
                    single_l = False
                    iord_line = None
                    nord_eff = data_dic[inst]['nord']
                    dim_exp_eff = data_dic[inst][vis]['dim_exp']
                 
            #Correct CCF
            else:
                single_l = True
                iord_line = 0
                nord_eff = data_dic[inst]['nord']
                dim_exp_eff = data_dic[inst][vis]['dim_exp']
                
            #------------------------------------------
            #Trend corrections
            if cond_trend:
                print('           Correcting for trends')
                corr_prop = detrend_prof_dic['prop'][inst][vis]
            
                #Default orders for SNR
                if (inst in detrend_prof_dic['SNRorders']):SNRorders_inst = detrend_prof_dic['SNRorders'][inst]
                else:SNRorders_inst = {'HARPS':[49],'HARPN':[46],'ESPRESSO_MR':[39],'ESPRESSO':[102,103],'CARMENES_VIS':[40],
                                       'NIRPS_HA':[57],'NIRPS_HE':[57],    #H band, 1.63 mic, order not affected by tellurics thus stable for SNR measurement
                                       'EXPRES':[14]}[inst]                #562 nm

                #Initialize corrections
                #    - FWHM corrections can only be applied to single lines
                corr_list = list(corr_prop.keys())
                prop_corr = [corr_loc.split('_')[0] for corr_loc in corr_list]
                var_corr = [corr_loc.split('_')[1] for corr_loc in corr_list]
             
                if np.any([corr_loc=='RV' for corr_loc in prop_corr]):corr_RV=True
                else:corr_RV=False
                if np.any([corr_loc=='ctrst' for corr_loc in prop_corr]):
                    corr_ctrst=True
                    cond_def_cont_all  = np.zeros(data_vis['dim_ord'],dtype=bool)
                else:corr_ctrst=False            
                if np.any([corr_loc=='FWHM' for corr_loc in prop_corr]):
                    if single_l:corr_FWHM=True
                    else:stop('WARNING: FWHM correction cannot be applied to multiple spectral lines')
                else: corr_FWHM=False
                
                #Sanity check
                if corr_FWHM and data_dic['DI']['sysvel'][inst][vis]==0.:
                    print('         WARNING: sysvel = 0 km/s. The correct sysvel must be defined for the FWHM correction to be applied.')
                
                #Defining corrections
                glob_corr_ctrst = np.ones(data_vis['n_in_visit'],dtype=float)
                glob_corr_FWHM = np.ones(data_vis['n_in_visit'],dtype=float)
                glob_corr_RV = np.zeros(data_vis['n_in_visit'],dtype=float)
                for iexp in range(data_vis['n_in_visit']):
    
                    #SNR in chosen spectral orders
                    #    - indexes relative to original orders
                    if np.any(['snr' in corr_loc for corr_loc in var_corr]):
                        SNR_exp_ord = data_prop_vis['SNRs'][iexp] 
                        SNR_exp =np.mean(SNR_exp_ord[SNRorders_inst])
                        if np.any([corr_loc=='snrQ' for corr_loc in var_corr]):SNRQ_exp =np.sqrt(np.sum(SNR_exp_ord[SNRorders_inst]**2.))                 
                        
                    #Orbital phase
                    if np.any([corr_loc=='phase' for corr_loc in var_corr]):
                        cen_ph_exp=coord_dic[inst][vis][pl_loc]['cen_ph'][iexp] 
                 
                    #Airmass and activity indexes
                    env_prop = {}
                    if np.any([corr_loc=='AM' for corr_loc in var_corr]):
                        env_prop['AM']=data_prop_vis['AM'][iexp]     
                    for key in data_vis['act_idx']:
                        if np.any([corr_loc==key for corr_loc in var_corr]):
                            env_prop[key]=data_prop_vis[key][iexp,0]               
                
                    #----------------------------------------------------------------------------------------                    
                    #Polynomial corrections for contrast variation
                    #    - only the modulation around the constant value is corrected for
                    #---------------------------------------------------------------------------------------- 
                    if corr_ctrst:            
                        
                        #Upload latest processed DI data
                        data_exp = dataload_npz(data_vis['proc_DI_data_paths']+str(iexp))
  
                        #Continuum for single line
                        if single_l:
                          
                            #Initializing continuum ranges in the input rest frame to defined pixels over requested ranges 
                            for bd_int in data_dic['DI']['cont_range'][inst][iord_line]:
                                cond_def_cont_all[iexp] |= (data_exp['edge_bins'][0,0:-1]>=bd_int[0]) & (data_exp['edge_bins'][0,1:]<=bd_int[1])    
                            cond_def_cont_all[iexp] &= data_exp['cond_def'][0]
                            
                            #Exclusion of planetary ranges
                            if ('DI_prof' in data_dic['Atm']['no_plrange']) and (iexp in data_dic['Atm'][inst][vis]['iexp_no_plrange']):  
                                cond_def_cont_all[iexp] &= excl_plrange(data_exp['cond_def'][0],data_dic['Atm'][inst][vis]['exclu_range_input'],iexp,data_exp['edge_bins'][0],data_dic[inst][vis]['type'])[0]
    
                        #------------------------
    
                        #With phase   
                        if 'ctrst_phase' in corr_list:glob_corr_ctrst[iexp] = detrend_prof_gen(corr_prop['ctrst_phase'],cen_ph_exp,glob_corr_ctrst[iexp],'modul')
                        
                        #With SNR           
                        if 'ctrst_snr' in corr_list:glob_corr_ctrst[iexp] = detrend_prof_gen(corr_prop['ctrst_snr'],SNR_exp,glob_corr_ctrst[iexp],'modul')
                        elif 'ctrst_snrQ' in corr_list:glob_corr_ctrst[iexp] = detrend_prof_gen(corr_prop['ctrst_snrQ'],SNRQ_exp,glob_corr_ctrst[iexp],'modul')
    
                        #With airmass and indexes
                        for key in env_prop:
                            if 'ctrst_'+key in corr_list:glob_corr_ctrst[iexp] = detrend_prof_gen(corr_prop['ctrst_'+key],env_prop[key],glob_corr_ctrst[iexp],'modul')
                            
                    #----------------------------------------------------------------------------------------                    
                    #Polynomial corrections for FWHM variation
                    #    - only the modulation around the constant value is corrected for
                    #---------------------------------------------------------------------------------------- 
                    if corr_FWHM:                   
    
                        #With phase   
                        if 'FWHM_phase' in corr_list:glob_corr_FWHM[iexp] = detrend_prof_gen(corr_prop['FWHM_phase'],cen_ph_exp,glob_corr_FWHM[iexp],'modul')
                          
                        #With SNR           
                        if 'FWHM_snr' in corr_list:glob_corr_FWHM[iexp] = detrend_prof_gen(corr_prop['FWHM_snr'],SNR_exp,glob_corr_FWHM[iexp],'modul')
                        elif 'FWHM_snrQ' in corr_list:glob_corr_FWHM[iexp] = detrend_prof_gen(corr_prop['FWHM_snrQ'],SNRQ_exp,glob_corr_FWHM[iexp],'modul')
                        
                        #With airmass and indexes
                        for key in env_prop:
                            if 'FWHM_'+key in corr_list:glob_corr_FWHM[iexp] = detrend_prof_gen(corr_prop['FWHM_'+key],env_prop[key],glob_corr_FWHM[iexp],'modul')
    
                    #----------------------------------------------------------------------------------------                 
                    #Polynomial correction for deviation to Keplerian RV curve (km/s)
                    #    - the full polynomial variation is corrected for
                    #----------------------------------------------------------------------------------------       
                    if corr_RV:    
                        
                        #With phase
                        if 'RV_phase' in corr_list:
                            glob_corr_RV[iexp] = detrend_prof_gen(corr_prop['RV_phase'],cen_ph_exp,glob_corr_RV[iexp],'add')
                            
                        #With SNR
                        if 'RV_snr' in corr_list:
                            glob_corr_RV[iexp] = detrend_prof_gen(corr_prop['RV_snr'],SNR_exp,glob_corr_RV[iexp],'add')
    
                        elif 'RV_snrQ' in corr_list:
                            glob_corr_RV[iexp] = detrend_prof_gen(corr_prop['RV_snrQ'],SNRQ_exp,glob_corr_RV[iexp],'add')
    
                        #With airmass and indexes
                        for key in env_prop:
                            if 'RV_'+key in corr_list:glob_corr_RV[iexp] = detrend_prof_gen(corr_prop['RV_'+key],env_prop[key],glob_corr_RV[iexp],'add')
    
                #Normalizing around the mean
                glob_corr_ctrst/=np.mean(glob_corr_ctrst)                    
                glob_corr_FWHM/=np.mean(glob_corr_FWHM)                     
      
                #Continuum for contrast corrections
                if corr_ctrst:
                
                    #Continuum common to all input single line profiles 
                    if single_l:
                        cond_cont_com  = np.all(cond_def_cont_all,axis=0)
                        if np.sum(cond_cont_com)==0.:stop('No pixels in common continuum')
                        cont_func_dic= None
                    
                    #Continuum of stellar spectrum
                    else:
                        cond_cont_com=None
                        cont_func_dic = dataload_npz(gen_dic['save_data_dir']+'Stellar_cont_DI/'+inst+'_'+vis+'/St_cont')['cont_func_dic']

            #------------------------------------------
            #Custom correction
            if cond_custom:
                print('           Applying custom correction')
                cust_corr_RV = np.zeros(data_vis['n_in_visit'],dtype=float)
                corr_data = ((pd.read_csv('/Users/bourrier/Travaux/Exoplanet_systems/Divers/55_cancri/RM/RMR/Data/EXPRES/220728_55CnceTransit_detrend.csv')).values).T
                if vis=='20220131':JD_corr = corr_data[1]-2400000. 
                if vis=='20220406':JD_corr = corr_data[1]-2400000. 
                RV_corr = corr_data[2]*1e-3   
                if vis=='20220406':RV_corr*=-1. 
                for iexp in range(data_vis['n_in_visit']):
                    icorr = closest(JD_corr,coord_dic[inst][vis]['bjd'][iexp])
                    cust_corr_RV[iexp]=RV_corr[icorr]
         
            #------------------------------------------
            #PC corrections
            if cond_PC:
                print('           Correcting for PC')
                
                #PCA results
                pca_results = np.load(detrend_prof_dic['PC_model'][inst][vis]['all'],allow_pickle=True)['data'].item() 

                #Joint intrinsic fit
                if 'in' in detrend_prof_dic['PC_model'][inst][vis]:
                    jointfit_results = np.load(detrend_prof_dic['PC_model'][inst][vis]['in'],allow_pickle=True)['data'].item() 
                    if pca_results['n_pc']!=jointfit_results['n_pc'][inst][vis]:stop('Number of fitted PC must match')
                else:jointfit_results=None
                
                #PC profiles fitted to the residual and intrinsic data
                eig_res_matr = pca_results['eig_res_matr'][0:pca_results['n_pc']]
                
                #PC profiles selected to generate the model
                if (inst in detrend_prof_dic['idx_PC']) and (vis in detrend_prof_dic['idx_PC'][inst]):idx_pc = detrend_prof_dic['idx_PC'][inst][vis]
                else:idx_pc = range(pca_results['n_pc'])
                
            #----------------------------------------------------------------------------------------               

            #Resample aligned profiles on the common visit table if relevant
            if (data_vis['comm_sp_tab']):
                data_com = dataload_npz(data_vis['proc_com_data_paths'])
                cen_bins_resamp, edge_bins_resamp = data_com['cen_bins'],data_com['edge_bins']
            else:cen_bins_resamp, edge_bins_resamp = None,None 

            #Correct each exposure  
            proc_DI_data_paths_new = gen_dic['save_data_dir']+'Detrend_prof/'+inst+'_'+vis+'_'                          
            for iexp in range(data_vis['n_in_visit']):
        
                #Upload latest processed DI data
                data_exp = dataload_npz(data_vis['proc_DI_data_paths']+str(iexp))                  
 
                #Switch spectrum into RV space
                #    - upon radial velocity table associated with chosen transition
                if ('spec' in data_dic[inst][vis]['type']) and (detrend_prof_dic['line_trans'] is not None):
                    data_corr = {'flux' : np.array([data_exp['flux'][iord_line]]),
                                 'cov' : np.array([data_exp['cov'][iord_line]]),
                                 'edge_bins' : np.array([c_light*( (data_exp['edge_bins'][iord_line]/detrend_prof_dic['line_trans']) - 1.)]),
                                 'cen_bins' : np.array([c_light*( (data_exp['cen_bins'][iord_line]/detrend_prof_dic['line_trans']) - 1.)])}
 
                #Correct CCF or full spectrum
                else:data_corr = data_exp
                 
                #---------------------------------  
                #Trend corrections
                if cond_trend:

                    #Correcting for RV variation    
                    if corr_RV:
                        data_corr = align_data(data_corr,data_dic[inst][vis]['type'],nord_eff,dim_exp_eff,gen_dic['resamp_mode'],cen_bins_resamp, edge_bins_resamp,glob_corr_RV[iexp],1./gen_specdopshift(glob_corr_RV[iexp]))

                    #Correcting for contrast variation 
                    if corr_ctrst:                 
                        data_corr = detrend_prof_ctrst(data_corr,single_l,nord_eff,data_dic[inst][vis]['nspec'],glob_corr_ctrst[iexp],cond_cont_com,cont_func_dic,coord_dic[inst][vis]['RV_star_stelCDM'][iexp],data_dic['DI']['sysvel'][inst][vis],
                                                         gen_dic['save_data_dir']+'Scaled_data/'+inst+'_'+vis+'_scaling_'+str(iexp),coord_dic[inst][vis]['t_dur'][iexp])

                    #Correcting for FWHM variation 
                    #    - for CCF and single spectral lines 
                    if corr_FWHM and single_l:                          
                        glob_corr=np.repeat(glob_corr_FWHM[iexp],data_vis['nspec']+1)
                        data_corr = detrend_prof_FWHM(data_corr,glob_corr,coord_dic[inst][vis]['RV_star_solCDM'][iexp],gen_dic['resamp_mode'],comm_sp_tab=data_vis['comm_sp_tab'])

                #---------------------------------  
                #Custom correction
                if cond_custom:
                    data_corr['cen_bins'][0] -= cust_corr_RV[iexp]
                    data_corr['edge_bins'][0] -= cust_corr_RV[iexp]
      
                #---------------------------------      
                #PC correction
                #    - see PCA module for details
                if cond_PC: 
                    i_in = gen_dic[inst][vis]['idx_exp2in'][iexp]
                    
                    #Correct exposure if included in PCA module or joint intrinsic fit
                    if (iexp in pca_results['idx_corr']) or ((jointfit_results is not None) and (i_in in jointfit_results['idx_in_fit'][inst][vis])):
        
                        #Upload flux scaling 
                        data_scaling = dataload_npz(gen_dic['save_data_dir']+'Scaled_data/'+inst+'_'+vis+'_scaling_'+str(iexp))

                        #Switch PCA spectrum into RV space
                        if ('spec' in data_dic[inst][vis]['type']) and (detrend_prof_dic['line_trans'] is not None):
                            bins_edge_PCA = c_light*( (pca_results['edge_bins'][0]/detrend_prof_dic['line_trans']) - 1.)  
                        else:bins_edge_PCA = pca_results['edge_bins'][0]                           

                        #Residual PC model from intrinsic profile fit
                        pc_mod_exp = np.zeros(data_vis['nspec'],dtype=float)
                        if ((jointfit_results is not None) and (i_in in jointfit_results['idx_in_fit'][inst][vis])):
                     
                            #Intrinsic PC model
                            for i_pc in idx_pc:pc_mod_exp+=jointfit_results['p_final']['aPC_idxin'+str(i_in)+'_ord'+str(i_pc)+'__IS'+inst+'_VS'+vis]*eig_res_matr[i_pc]
                            
                            #Scaling to level of residual data 
                            #    - see PCA module, the intrinsic PC model corresponds to :
                            # Pfit = -Pert(w,t,v)*LC_theo(band,t)*Cref(band,v)/(1 - LC_theo(band,t)) 
                            #      which can be scaled to the level of the disk-integrated flux by
                            # globF(v,t)*(1 - LC_theo(band,t))/LC_theo(band,t)
                            LC_theo_exp = 1. - data_scaling['loc_flux_scaling'][iexp](bins_edge_PCA)
                            pc_mod_exp*= (1. - LC_theo_exp)/LC_theo_exp
                      
                        #Residual PC model from PCA analysis
                        elif (iexp in pca_results['idx_corr']):
                            for i_pc in idx_pc:pc_mod_exp+=pca_results['p_final']['aPC_idx'+str(iexp)+'_ord'+str(i_pc)+'__IS'+inst+'_VS'+vis]*eig_res_matr[i_pc]
               
                        #Save residual model for plotting purposes
                        if plot_dic['map_pca_prof']!='':
                            np.savez_compressed(gen_dic['save_data_dir']+'PCA_results/'+inst+'_'+vis+'_model'+str(iexp) ,data = {'flux':np.array([pc_mod_exp]),'edge_bins' :np.array([bins_edge_PCA]),'cond_def':np.ones(data_dic[inst][vis]['dim_exp'],dtype=bool) },allow_pickle=True)
                
                        #Scale to the level of disk-integrated flux density
                        pc_mod_exp*=data_scaling['glob_flux_scaling'][iexp]
  
                        #Temporary exposure table centered in star rest frame
                        edge_bins_shift = data_corr['edge_bins'][0] - coord_dic[inst][vis]['RV_star_solCDM'][iexp] 
                      
                        #PC model interpolated over exposure table
                        pc_mod_exp = bind.resampling(edge_bins_shift,bins_edge_PCA, pc_mod_exp, kind=gen_dic['resamp_mode'])  
                     
                        #Correction
                        data_corr['flux'][0]+=pc_mod_exp
    
                #-----------------------------------------------------
                #Saving corrected data
                #----------------------------------------------------- 
    
                #Switch spectrum back into wavelength space
                #    - wave = wave_ref*(1 + rv/c)
                if ('spec' in data_dic[inst][vis]['type']) and (detrend_prof_dic['line_trans'] is not None):
                    data_exp['flux'][iord_line] = data_corr['flux'][0]
                    data_exp['cov'][iord_line] = data_corr['cov'][0]
                    data_exp['edge_bins'][iord_line] = detrend_prof_dic['line_trans']*gen_specdopshift(data_corr['edge_bins'][0])     
                    data_exp['cen_bins'][iord_line]  = detrend_prof_dic['line_trans']*gen_specdopshift(data_corr['cen_bins'][0])                      
                else:data_exp = data_corr
    
                #Updating defined bins
                data_exp['cond_def'] = ~np.isnan(data_exp['flux'])
    
                #Saving corrected data and updating paths 
                np.savez_compressed(proc_DI_data_paths_new+str(iexp) ,data = data_exp,allow_pickle=True)
            
            ### end of exposure    
            data_vis['proc_DI_data_paths'] = proc_DI_data_paths_new
         
        else: 
            #Updating path to processed data and checking it has been calculated
            data_vis['proc_DI_data_paths'] = gen_dic['save_data_dir']+'Detrend_prof/'+inst+'_'+vis+'_' 
            check_data({'path':data_vis['proc_DI_data_paths']+str(0)})

    return None


def detrend_prof_gen(corr_prop_in,var,corr_in,mode):
    """**Detrending: generic model.**
    
    Calculates polynomial or sinusoidal model, multiplicative or cumulative.

    Args:
        TBD
    
    Returns:
        TBD
    
    """          
    corr_out = deepcopy(corr_in)
    
    #Modulated variation
    if mode=='modul':
        if ('sin' in corr_prop_in):
            loc_corr = (1. + corr_prop_in['sin'][0]*np.sin(2*np.pi*((var-corr_prop_in['sin'][1])/corr_prop_in['sin'][2])))
            corr_out*=loc_corr
        if ('pol' in corr_prop_in) and (len(corr_prop_in['pol'])>0):
            loc_corr = 1.
            for ideg in range(1,len(corr_prop_in['pol'])+1):loc_corr+=corr_prop_in['pol'][ideg-1]*(var**ideg)
            corr_out*=loc_corr

    #Cumulative variation
    elif mode=='add':
        if ('sin' in corr_prop_in):
            loc_corr = corr_prop_in['sin'][0]*np.sin(2*np.pi*((var-corr_prop_in['sin'][1])/corr_prop_in['sin'][2]))
            corr_out+=loc_corr
        if ('pol' in corr_prop_in) and (len(corr_prop_in['pol'])>0):
            loc_corr = 0.
            for ideg in range(1,len(corr_prop_in['pol'])+1):loc_corr+=corr_prop_in['pol'][ideg-1]*(var**ideg)
            corr_out+=loc_corr
        
    return corr_out



def detrend_prof_ctrst(data_corr,single_l,nord_eff,nspec,glob_corr_ctrst_exp,cond_cont_com,cont_func_dic,rv_star_stelCDM,rv_stelCDM_solCDM,scaled_DI_data_paths_exp,t_dur_exp):
    r"""**Detrending: contrast model.**
    
    Corrects line profile for contrast variations.
    
    Line profile is temporarily set to a null continuum so that its contrast can be 'stretched' by the correction factor :math:`\alpha`
    
    .. math::     
       C_\mathrm{corr} &= ( ((F/F_\mathrm{cont})-1)/\alpha  + 1) F_\mathrm{cont}   \\
                       &= (F-F_\mathrm{cont})/\alpha  + F_\mathrm{cont}    \\
                       &= (F/\alpha) + F_\mathrm{cont} (1 - (1/\alpha)) 
                
    We assume that the contrast variation arises from a modification of the measured profile akin to a change in LSF. 
    All profiles (including in-transit) are thus set to a common vertical scale in which the line contrast is equivalent, by normalizing them with the stellar continuum x the profile global flux level.   
    Both continuum and level must have been defined from a first processing of the data.   

    Args:
        TBD
    
    Returns:
        TBD
    
    """  
    
    #Correcting single line profile   
    if single_l:

        #CCF continuum flux
        #    - defined in the same way as when fitting CCFs, to be consistent with the definition of the contrast
        cont_CCF=np.mean(data_corr['flux'][0,cond_cont_com])
                 
        #Correcting for contrast variation
        #    - the covariance is only modified by the correction factor
        corr_val = np.repeat(glob_corr_ctrst_exp,nspec) 
        data_corr['flux'][0],data_corr['cov'][0] = bind.mul_array(data_corr['flux'][0],data_corr['cov'][0],1./corr_val)        
        data_corr['flux'][0] += cont_CCF*(1. - (1./corr_val))
        
    #Correcting full spectrum
    else:    
        
        #Align spectra from solar barycentric (receiver) to star (source) rest frame
        #      see gen_specdopshift() :
        # w_source = w_receiver / (1+ (rv[s/r]/c))
        # w_star =  w_solbar / ((1+ rv[star/starbar]/c))* (1+ rv[starbar/solbar]/c)))
        #    - the profile is only temporarily shifted and does not need resampling
        cen_bins_rest = data_corr['cen_bins']/(gen_specdopshift(rv_star_stelCDM)*gen_specdopshift(rv_stelCDM_solCDM)) 

        #Processing each exposure                   
        glob_flux_scaling = dataload_npz(scaled_DI_data_paths_exp)['glob_flux_scaling']   
        glob_sc = np.repeat(1./(t_dur_exp*glob_flux_scaling),nspec) 
        for iord in range(nord_eff):

            #Set spectrum to common vertical range (set to common flux level and correct spectrum for stellar continuum), stretch/correct/unstretch, reset to original range
            #    - broadband flux scaling is not applied to maintain all profiles to the same flux balance
            comm_sc = glob_sc/cont_func_dic(cen_bins_rest[iord])
            data_corr['flux'][iord],data_corr['cov'][iord] = bind.mul_array(data_corr['flux'][iord],data_corr['cov'][iord],comm_sc/glob_corr_ctrst_exp)
            data_corr['flux'][iord] += (1. - (1./glob_corr_ctrst_exp))
            data_corr['flux'][iord],data_corr['cov'][iord] = bind.mul_array(data_corr['flux'][iord],data_corr['cov'][iord],1./comm_sc)

    return data_corr



def detrend_prof_FWHM(data_corr,FWHM_corr,RV_star_solCDM,resamp_mode,comm_sp_tab=False):
    r"""**Detrending: FWHM model.**
    
    Corrects line profile for FWHM variations.
    
    Performs a stretch of the spectral tables in rv space. 
    Modifying the width of a line profile to get FWHM/corr is equivalent to redefining the line profile on its rv table divided by the FWHM correction.
    Corrected profiles are resampled on the common visit table if relevant (if a common table is used then all original tables point toward this common table). 
    This operation must be performed on velocity tables symmetrical with respect to the line center, ie for lines that have been aligned on the null velocity.  
    This means the RV correction must be performed first, to determine and fix the correct systemic rv. 

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    #Temporary spectral table
    #    - stretched by the FWHM correction
    edge_bins_shift = data_corr['edge_bins'][0] - RV_star_solCDM 
    edge_bins_stretch = edge_bins_shift/FWHM_corr
    
    #Resampling on common table
    if comm_sp_tab:
        data_corr['flux'][0],data_corr['cov'][0] = bind.resampling(edge_bins_shift, edge_bins_stretch, data_corr['flux'][0] , cov = data_corr['cov'][0], kind=resamp_mode)   
        data_corr['cond_def'][0] = ~np.isnan(data_corr['flux'][0])
    else:
        data_corr['edge_bins'][0]=edge_bins_stretch+RV_star_solCDM
        data_corr['cen_bins'][0]=0.5*(data_corr['edge_bins'][0][0:-1]+data_corr['edge_bins'][0][1::])
        
    return data_corr





def pc_analysis(gen_dic,data_dic,inst,vis,data_prop,coord_dic):
    r"""**PC: main routine.**
    
    Applies Principal Components Analysis.
    
    We define the perturbed profiles as

    .. math::       
       F(w,t,v) = F_{\star}(w,v) + F_\mathrm{pert}(w,t,v)
          
    Where :math:`F_\mathrm{pert}` is the physical perturbation from the star, assumed to be roughly uniform over the stellar disk

    .. math:: 
       F_\mathrm{pert}(w,t,v) &\sim \sum_{k} I_\mathrm{pert}(w,t,v) LD_{k}(band) S_{k}  \\
                   &= I_\mathrm{pert}(w,t,v) \sum_{k} LD_{k}(band) S_{k}  \\ 
                   &= I_\mathrm{pert}(w,t,v) S_{\star,LD}(band)
                   
    The perturbation can be described as a linear combination of principal components `PC`, as                   

    .. math:: 
       F_\mathrm{pert}(w,t,v) = \sum_{k} a(t,v) PC(k,w)     
    
    We apply the correction to the raw disk-integrated profiles :math:`F(w,t,v) C_\mathrm{ref}(band,v) F_\mathrm{glob}(v,t)` in `detrend_prof()`, so that the correction model is

    .. math::     
       F_\mathrm{corr}^\mathrm{pca}(w,t,v) = F_\mathrm{pert}(w,t,v) C_\mathrm{ref}(band,v) F_\mathrm{glob}(v,t) 
     
    Residual profiles are calculated from the scaled spectra :math:`F_\mathrm{sc}(w,t) = F(w,t,v) C_\mathrm{ref}(w,v)`.  
    Out-of-transit they are defined as

    .. math:: 
       F_\mathrm{res}(w,t,v) &= F_{\star,sc}(w,v) - F_\mathrm{sc}(w,t,v)  \\
                    &= - F_\mathrm{pert}(w,t,v) C_\mathrm{ref}(band,v)              
    
    To which we fit :math:`F_\mathrm{res}(w,t,v) F_\mathrm{glob}(v,t)`     
    
    In-transit, outside of the planetary lines (see `proc_intr_data()`), they are defined as 

    .. math:: 
       F_\mathrm{intr}(w,t,v) &= F_\mathrm{res}(w,t,v)/(1 - LC_\mathrm{theo}(band,t))     \\
                    &= ( F_{\star,sc}(w,v) - F_\mathrm{sc}(w,t,v) )/(1 - LC_\mathrm{theo}(band,t))    \\
                    &= ( \sum_{k}(I_{k}(w,t,v) LD_{k}(band) S_{k})   - ( \sum_{k_\mathrm{unocc}}(I_{k}(w,t,v) LD_{k}(band) S_{k}) + \sum_{k_\mathrm{unocc}}(I_\mathrm{pert}(w,t,v) LD_{k}(band) S_{k}))) C_\mathrm{ref}(band,v)/(1 - LC_\mathrm{theo}(band,t))   \\
                    &= F_\mathrm{intr,pl}(w,t,v) - I_\mathrm{pert}(w,t,v) \sum_{k_\mathrm{unocc}}(LD_{k}(band) S_{k}) C_\mathrm{ref}(band,v)/(1 - LC_\mathrm{theo}(band,t))   \\
                    &= F_\mathrm{intr,pl}(w,t,v) - (F_\mathrm{pert}(w,t,v)/S_{\star,LD}(band)) (S_{\star,LD}(band) - LD(band,t) S_{p}(band,t)) C_\mathrm{ref}(band,v)/(1 - LC_\mathrm{theo}(band,t))   \\
                    &= F_\mathrm{intr,pl}(w,t,v) - F_\mathrm{pert}(w,t,v) (1 - (LD(band,t) S_{p}(band,t)/S_{\star,LD}(band))) C_\mathrm{ref}(band,v)/(1 - LC_\mathrm{theo}(band,t))                \\  
                    &= F_\mathrm{intr,pl}(w,t,v) - F_\mathrm{pert}(w,t,v) LC_\mathrm{theo}(band,t) C_\mathrm{ref}(band,v)/(1 - LC_\mathrm{theo}(band,t))                 
    
    Outside of the planet-occulted stellar lines the continuum of :math:`F_\mathrm{intr,pl}` is constant, and the continuum of :math:`F_\mathrm{pert}` is assumed to be null, so that

    .. math::                
       F_\mathrm{intr}(w,t,v) - F_\mathrm{intr}(cont,t,v) = - F_\mathrm{pert}(w,t,v) LC_\mathrm{theo}(band,t) C_\mathrm{ref}(band,v)/(1 - LC_\mathrm{theo}(band,t))
    
    We thus fit 
    
    .. math::
       (F_\mathrm{intr}(w,t,v) - F_\mathrm{intr}(cont,t,v)) F_\mathrm{glob}(v,t) (1 - LC_\mathrm{theo}(band,t))/LC_\mathrm{theo}(band,t)
             

    Args:
        TBD
    
    Returns:
        TBD
        
    """    
    if (inst in data_dic['PCA']['vis_list']) and ((vis in data_dic['PCA']['vis_list'][inst]) or (data_dic['PCA']['vis_list'][inst]=='all')):
        print('   > Applying PCA to OT residual profiles') 
        
        #Processing data
        if (gen_dic['calc_pca_ana']):
            print('         Calculating data')        
            data_vis=data_dic[inst][vis]
            if (data_vis['type']!='CCF'):stop('Only coded for CCFs')
            gen_vis = gen_dic[inst][vis]

            #Exposure selection
            idx_pca = gen_vis['idx_out']
            if (inst in data_dic['PCA']['idx_pca']) and (vis in data_dic['PCA']['idx_pca'][inst][vis]) and (len(data_dic['PCA']['idx_pca'][inst][vis])>0):
                idx_pca = np.intersect1d(data_dic['PCA']['idx_pca'][inst][vis],idx_pca)
            nexp_pca = len(idx_pca)
       
            #Analysis and fit range 
            #    - currently we assume that all exposures are defined over the same spectral table
            data_com = np.load(data_vis['proc_com_data_paths']+'.npz',allow_pickle=True)['data'].item()
            cond_out_fit_range = False   
            for bd_int in data_dic['PCA']['ana_range'][inst][vis]:
                cond_out_fit_range |= (data_com['edge_bins'][0,0:-1]>=bd_int[0]) & (data_com['edge_bins'][0,1:]<=bd_int[1])   

            #Store residual profiles as a matrix
            n_rep = 100
            full_res_matr = np.zeros([nexp_pca,data_vis['nspec']],dtype=float)*np.nan
            full_noise_matr = np.zeros([nexp_pca,data_vis['nspec'],n_rep],dtype=float)*np.nan
            cond_fit_range_matr = np.zeros([nexp_pca,data_vis['nspec']],dtype=bool)
            iexp2ipca = np.zeros(data_vis['n_in_visit'],dtype=int)-1
            isub_pca_pretr = []
            isub_pca_posttr = []
            for isub,iexp in enumerate(idx_pca):
                iexp2ipca[iexp] = isub
                if iexp in gen_vis['idx_pretr']:isub_pca_pretr+=[isub]
                else:isub_pca_posttr+=[isub]
                    
                #PCA based on out-of-transit data alone
                data_exp = np.load(data_vis['proc_Res_data_paths']+str(iexp)+'.npz',allow_pickle=True)['data'].item()
                cond_fit_range_matr[isub] = cond_out_fit_range & data_exp['cond_def'][0] 
                full_res_matr[isub] = data_exp['flux'][0]

                #Exclude planetary ranges
                if ('PCA_corr' in data_dic['Atm']['no_plrange']) and (iexp in data_dic['Atm'][inst][vis]['iexp_no_plrange']):
                    cond_in_pl = ~(np.ones(data_vis['nvel'],dtype=bool) & excl_plrange(data_exp['cond_def'][0],data_dic['Atm'][inst][vis]['exclu_range_star'],iexp,data_exp['edge_bins'][0],data_dic[inst][vis]['type'])[0])
                    cond_fit_range_matr[isub,cond_in_pl] = False

                #Generate 'noise' matrix 
                #    - we create n_rep realisations using the error array of current exposure
                err_exp = np.sqrt(data_exp['cov'][0][0])
                full_noise_matr[isub] = list(map(np.random.normal, np.zeros(data_vis['nspec']), err_exp, [n_rep] * data_vis['nspec']))

            #Reduce matrixes to analysis range and center each pixel on null level over time (ie the mean value over selected exposures is subtracted)
            cond_def_all = np.all(cond_fit_range_matr,axis=0)
            nspec_res_mat = np.sum(cond_def_all)
            n_pca_pretr = len(isub_pca_pretr)
            n_pca_posttr = len(isub_pca_posttr)
            nexp_pca = {'out':n_pca_pretr+n_pca_posttr}
            res_matr = {'out': full_res_matr[:,cond_def_all]}
            noise_matr = {'out': full_noise_matr[:,cond_def_all]}
            eig_vals = {}
            eig_val_noise = {}
            if n_pca_pretr>0:
                nexp_pca['pre'] =n_pca_pretr 
                res_matr['pre']=full_res_matr[isub_pca_pretr][:,cond_def_all]
                noise_matr['pre']=full_noise_matr[isub_pca_pretr][:,cond_def_all]
            if n_pca_posttr>0:
                nexp_pca['post'] =n_pca_posttr
                res_matr['post']=full_res_matr[isub_pca_posttr][:,cond_def_all]
                noise_matr['post']=full_noise_matr[isub_pca_posttr][:,cond_def_all]
            for key in res_matr:
                res_matr[key]-=np.nanmean(res_matr[key],axis=0)
                noise_matr[key]-=np.nanmean(noise_matr[key],axis=0)

                #Eigen decomposition
                #    - eigenvalues have dimension nexp, and are already sorted in descending order
                #      eigen_vecs have dimension nexp x nexp, where eig_vecs_matr[i,:] is the ith vector corresponding to the ith eigenvalue
                _, eig_vals_sqrt, eig_vecs_matr = np.linalg.svd(res_matr[key])
                eig_vals[key] = eig_vals_sqrt**2.
                eig_vecs_matr = eig_vecs_matr[0:nexp_pca[key]]

                #Store eigenvectors into null matrix with same dimension as original data
                if key=='out':
                    eig_res_matr = np.zeros([nexp_pca[key],data_vis['nspec']],dtype=float)
                    eig_res_matr[:,cond_def_all] = eig_vecs_matr

                #Calculate eigenvalues for each realization and average them
                eig_val_noise[key] = np.zeros(nexp_pca[key],dtype=float)
                for iboot in range(n_rep):
                    _, eig_vals_noise_sqrt, _ = np.linalg.svd(noise_matr[key][:,:,iboot])
                    eig_val_noise[key]+= eig_vals_noise_sqrt**2.
                eig_val_noise[key]/=n_rep
            
            #--------------------------------------------------------------------
            #Fit PC to all exposures
            fixed_args = {'use_cov':False,'suff':'__IS'+inst+'_VS'+vis}
            
            #PC to use
            fixed_args['n_pc'] = data_dic['PCA']['n_pc'][inst][vis]
            eig_res_matr_fit = eig_res_matr[0:fixed_args['n_pc']]
                
            #Exposures to correct
            idx_corr = np.arange(data_vis['n_in_visit'])
            if (inst in data_dic['PCA']['idx_corr']) and (vis in data_dic['PCA']['idx_corr'][inst][vis]) and (len(data_dic['PCA']['idx_corr'][inst][vis])>0):
                idx_corr = np.intersect1d(data_dic['PCA']['idx_corr'][inst][vis],idx_corr)

            #Fit range
            if (inst not in data_dic['PCA']['fit_range']) or (vis not in data_dic['PCA']['fit_range'][inst]):
                fit_range = data_dic['PCA']['ana_range'][inst][vis]
            else:fit_range=data_dic['PCA']['fit_range'][inst][vis]

            #Process selected exposures
            rms_full_res_matr = np.zeros([2,data_vis['n_in_visit']],dtype=float)*np.nan
            corr_res_matr = np.zeros([nexp_pca['out'],nspec_res_mat],dtype=float)
            chi2null_tab = np.zeros(data_vis['n_in_visit'],dtype=float)*np.nan
            chi2_tab = np.zeros(data_vis['n_in_visit'],dtype=float)*np.nan
            BIC_tab = np.zeros(data_vis['n_in_visit'],dtype=float)*np.nan
            nfit_tab = np.zeros(data_vis['n_in_visit'],dtype=float)*np.nan
            p_final = {}   
            for isub,iexp in enumerate(idx_corr): 

                #Initialise fit parameters 
                p_start = Parameters()
                for i_pc in range(fixed_args['n_pc']):p_start.add_many(('aPC_idx'+str(iexp)+'_ord'+str(i_pc)+'__IS'+inst+'_VS'+vis, 0.,  True, None,None,  None))                
                fixed_args['iexp'] = str(iexp)

                #Fitted range and continuum level
                #    - see explanations in function comments
                data_exp = np.load(data_vis['proc_Res_data_paths']+str(iexp)+'.npz',allow_pickle=True)['data'].item() 
                cond_fit_range = False
                for bd_int in fit_range:cond_fit_range |= (data_exp['edge_bins'][0,0:-1]>=bd_int[0]) & (data_exp['edge_bins'][0,1:]<=bd_int[1])  
                cond_fit_range &=  data_exp['cond_def'][0]

                #Fit PCs
                fixed_args['idx_fit'] = np_where1D(cond_fit_range)
                fixed_args['eig_res_matr'] = eig_res_matr_fit[:,cond_fit_range]
                result, merit ,p_best = call_lmfit(p_start,data_exp['cen_bins'][0],data_exp['flux'][0],data_exp['cov'][0],pc_model,verbose=True,fixed_args=fixed_args)
                for key in p_best:p_final[key] = p_best[key].value

                #BIC for null model (=chi2) and best-fit model
                flux_fit = data_exp['flux'][0,cond_fit_range]
                cov_fit = data_exp['cov'][0][:,cond_fit_range] 
                chi2null_tab[iexp] = np.sum( flux_fit**2. / cov_fit[0])
                nfit_tab[iexp] = np.sum(cond_fit_range)
                chi2_tab[iexp] = merit['chi2']
                BIC_tab[iexp] = merit['BIC']

                #Measure RMS pre/post correction on fitted residual profile
                rms_full_res_matr[0,iexp] = np.std(flux_fit)
                rms_full_res_matr[1,iexp] = np.std(flux_fit - merit['fit'])
                
                #Measure RMS pre/post correction on defined residual profile used for the PCA
                if iexp2ipca[iexp]>-1:
                    isub_pca = iexp2ipca[iexp]
                    res_exp = full_res_matr[isub_pca,cond_def_all]
                  
                    fixed_args['eig_res_matr'] = eig_res_matr_fit[:,cond_def_all]
                    corr_res_exp = res_exp - pc_model(p_best,data_com['cen_bins'][0,cond_def_all],args = fixed_args)                  
                
                    #Store corrected residual profile used for the PCA
                    #    - OT profiles are fully defined over 'cond_def_all'
                    corr_res_matr[isub_pca] = corr_res_exp

            #--------------------------------------------------------------------
            #Quality assessment
            #    - apply FFT to defined residual profiles used for the PCA
            #    - we separate pre- and post-transit exposures to avoid introducing correlation, if the noise is not white

            #Max fft of residual profiles
            #    - mean-centering does not change the result   
            fft_pre = np.nan if n_pca_pretr==0. else np.max(np.abs(fft2(res_matr['pre']))**2.)
            fft_post = np.nan if n_pca_posttr==0. else np.max(np.abs(fft2(res_matr['post']))**2.)
            max_fft2_res_matr = [ fft_pre , fft_post ]
                            
            #Max fft of corrected residual profiles used for PCA
            fft_pre = np.nan if n_pca_pretr==0. else np.max(np.abs(fft2(corr_res_matr[isub_pca_pretr]))**2.)
            fft_post = np.nan if n_pca_posttr==0. else np.max(np.abs(fft2(corr_res_matr[isub_pca_posttr]))**2.)
            max_fft2_corr_res_matr = [ fft_pre , fft_post ]
            
            #Distribution of residuals, post-correction
            hist_corr_res_pre = np.histogram(corr_res_matr[isub_pca_pretr],bins=data_dic['PCA']['nbins'],density=True)
            std_corr_res_pre = np.std(corr_res_matr[isub_pca_pretr])
            hist_corr_res_post = np.histogram(corr_res_matr[isub_pca_posttr],bins=data_dic['PCA']['nbins'],density=True)
            std_corr_res_post = np.std(corr_res_matr[isub_pca_posttr])
            hist_corr_res = np.histogram(corr_res_matr,bins=data_dic['PCA']['nbins'],density=True)
            std_corr_res = np.std(corr_res_matr)

            #Loop over velocities
            fft1D_res_matr = np.zeros([2,nspec_res_mat],dtype=float)*np.nan
            fft1D_corr_res_matr = np.zeros([2,nspec_res_mat],dtype=float)*np.nan
            fft1D_boot_res_matr = np.zeros([2,nspec_res_mat],dtype=float)
            if n_pca_pretr>0:boot_res_matr_pretr = np.zeros([n_pca_pretr,nspec_res_mat,data_dic['PCA']['nboot']],dtype=float) 
            if n_pca_posttr>0:boot_res_matr_posttr = np.zeros([n_pca_posttr,nspec_res_mat,data_dic['PCA']['nboot']],dtype=float) 
            for ipix in range(nspec_res_mat):
                
                #Pre-transit data
                if n_pca_pretr>0:
                    
                    #FFT on current velocity column
                    fft1D_res_matr[0,ipix] = np.max(np.abs(fft(res_matr['pre'][:,ipix]))**2.)
                    fft1D_corr_res_matr[0,ipix] = np.max(np.abs(fft(corr_res_matr[isub_pca_pretr,ipix]))**2.)
                    for iboot in range(data_dic['PCA']['nboot']):
                        boot_res_matr_pretr[:,ipix,iboot] = np.random.choice(corr_res_matr[isub_pca_pretr,ipix],n_pca_pretr,replace=False)                        
                        fft1D_boot_res_matr[0,ipix] += np.max(np.abs(fft(boot_res_matr_pretr[:,ipix,iboot]))**2.)
                    fft1D_boot_res_matr[0,ipix]/=n_pca_pretr

                #Post-transit data
                if n_pca_posttr>0:

                    #FFT on current velocity column
                    fft1D_res_matr[1,ipix] = np.max(np.abs(fft(res_matr['post'][:,ipix]))**2.)
                    fft1D_corr_res_matr[1,ipix] = np.max(np.abs(fft(corr_res_matr[isub_pca_posttr,ipix]))**2.)
                    for iboot in range(data_dic['PCA']['nboot']):
                        boot_res_matr_posttr[:,ipix,iboot] = np.random.choice(corr_res_matr[isub_pca_posttr,ipix],n_pca_posttr,replace=False) 
                        fft1D_boot_res_matr[1,ipix] += np.max(np.abs(fft(boot_res_matr_posttr[:,ipix,iboot]))**2.)
                    fft1D_boot_res_matr[1,ipix]/=n_pca_posttr

            #Max fft of corrected residual profiles used for PCA, bootstrapped                        
            max_fft2_boot_res_matr = np.zeros(2,dtype=float)
            for iboot in range(data_dic['PCA']['nboot']):
                if n_pca_pretr>0:max_fft2_boot_res_matr[0] +=  np.max(np.abs(fft2(boot_res_matr_pretr[:,:,iboot]))**2.)
                if n_pca_posttr>0:max_fft2_boot_res_matr[1] +=  np.max(np.abs(fft2(boot_res_matr_posttr[:,:,iboot]))**2.)
            if n_pca_pretr>0:
                max_fft2_boot_res_matr[0]/=n_pca_pretr
            else:max_fft2_boot_res_matr[0] = np.nan
            if n_pca_posttr>0:max_fft2_boot_res_matr[1]/=n_pca_posttr
            else:max_fft2_boot_res_matr[1] = np.nan                

            #--------------------------------------------------------------------

            #Save PCA results
            data_save = {'eig_val_res':eig_vals,'eig_val_noise':eig_val_noise,'eig_res_matr':eig_res_matr,'idx_pca':idx_pca,'isub_pca':np_where1D(iexp2ipca>-1),'idx_corr':idx_corr,'cen_bins':data_com['cen_bins'][0],
                         'n_pc':fixed_args['n_pc'],'rms_full_res_matr':rms_full_res_matr,'BIC_tab':BIC_tab,'chi2_tab':chi2_tab,'chi2null_tab':chi2null_tab,'nfit_tab':nfit_tab,
                         'edge_bins' : data_com['edge_bins'],'p_final':p_final,
                         'max_fft2_res_matr':max_fft2_res_matr,'max_fft2_corr_res_matr':max_fft2_corr_res_matr,'max_fft2_boot_res_matr':max_fft2_boot_res_matr,
                         'cen_bins_res_mat':data_com['cen_bins'][0,cond_def_all],'fft1D_res_matr':fft1D_res_matr,'fft1D_corr_res_matr':fft1D_corr_res_matr,'fft1D_boot_res_matr':fft1D_boot_res_matr,
                         'hist_corr_res_pre':hist_corr_res_pre,'hist_corr_res_post':hist_corr_res_post,'hist_corr_res':hist_corr_res,
                         'std_corr_res_pre':std_corr_res_pre,'std_corr_res_post':std_corr_res_post,'std_corr_res':std_corr_res}
            np.savez_compressed(gen_dic['save_data_dir']+'PCA_results/'+inst+'_'+vis,data=data_save,allow_pickle=True)   
    
        else:
            data_paths={'0':gen_dic['save_data_dir']+'PCA_results/'+inst+'_'+vis}
            check_data(data_paths)  

    return None

def pc_model(params,x,args = None):
    """**PC: model.**
    
    Calculates linear combination of Principal Components.

    Args:
        TBD
    
    Returns:
        TBD
    
    """     
    mod = np.zeros(len(x),dtype=float)
    for i_pc in range(args['n_pc']):mod+=params['aPC_idx'+args['iexp']+'_ord'+str(i_pc)+args['suff']]*args['eig_res_matr'][i_pc]
    return mod



   
def corr_length_determination(Res_data_vis,data_vis,scr_search,inst,vis,gen_dic):
    """**Correlation length.**
    
    Determine the spectral correlation length from out-of-transit residual profiles (so that values are spread around zero). 
    See method in Pont et al. 2006 and Bourrier et al. 2015 (tomography).

    Args:
        TBD
    
    Returns:
        TBD
    
    """  
    #Processing out-of-transit data
    Res_data_vis['corr_search']={
        'meas':np.zeros([data_vis['n_out_tr'],gen_dic['scr_srch_max_binwin']],dtype=float),
        'fit':np.zeros([data_vis['n_out_tr'],gen_dic['scr_srch_max_binwin']],dtype=float),
        'sig_corr':np.zeros([data_vis['n_out_tr']],dtype=float),
        'sig_uncorr':np.zeros([data_vis['n_out_tr']],dtype=float)
        }
    for isub,iexp in enumerate(gen_dic[inst][vis]['idx_out']):    
        
        #Upload data
        data_exp = np.load(data_vis['proc_Res_data_paths']+str(iexp)+'.npz',allow_pickle=True)['data'].item()
        n_pix = len(data_exp['flux'][0]) 
        
        #Calculate dispersion on binned points for each window bin size
        for ibin,nperbin in enumerate(gen_dic['scr_srch_nperbins']):
            
            #Number of possible positions for the sliding window, and corresponding positions
            #    - the last value for i_0 is n_win-1
            #                     for i_f is n_pts-1
            n_win=n_pix-nperbin+1.
            i_0_tab=np.arange(n_win)
            i_f_tab=nperbin-1.+np.arange(n_pix-nperbin+1.)                    
    
    #       #Placing sliding window successively with no overlap
    #       n_win=int(n_pix/nperbin)
    #       i_0_tab=np.arange(n_win)*nperbin
    #       i_f_tab=i_0_tab+nperbin-1.
    
            #Table of the mean residuals within a given window
            mean_tabs=np.zeros(n_win)
    
            #Sliding window along the whole time-series
            #    - steps are by one point, smaller than the sampling interval
            for idx,(i_0,i_f) in enumerate(zip(i_0_tab,i_f_tab)): 
      
                #Binning residuals within current window
                mean_tabs[idx]=np.mean(data_exp['flux'][i_0:i_f+1])
                          
            #Standard-deviation of the mean residuals, for current bin size 
            Res_data_vis['corr_search']['meas'][isub,ibin]=mean_tabs.std()
    
        #--------------------------------------------------------------
    
        #Measured values to fit
        xtofit = gen_dic['scr_srch_nperbins']
        ytofit = Res_data_vis['corr_search']['meas'][isub]
        covtofit = np.ones([1,len(xtofit)])
    
        #Guess
        #     -  sig_bin=1./np.power( np.power(sqrt(nbin)/sig_uncorr,4.) + np.power(1./sig_corr,4.) ,1./4.)
        # + for large bin size, uncorrelated noise should dominate: sig_bin ~  sig_uncorr / sqrt(nbin)
        # + for a bin size of one, we use the guess for the uncorrelated noise
        sig_uncorr= ytofit[-1]*np.sqrt(xtofit[-1])
        sig_corr = np.power( (1./np.power(ytofit[0],4.) ) - np.power(np.sqrt(1.)/sig_uncorr,4.) , -1./4. )       
    
        # Initialise fit parameters 
        #             (    Name,    Value,  Vary,   Min,   Max,  Expr)
        p_use = Parameters()
        p_use.add_many(( 'sig_uncorr', sig_uncorr,  True,None,None,  None),
                      (  'sig_corr', sig_corr,  True,None,None,  None))
    
        #Fitting
        result,merit,p_best = call_lmfit(p_use,xtofit,ytofit,covtofit,binned_stddev_fit,verbose=True)

        #Saving fit and its uncorrelated/correlated components 
        Res_data_vis['corr_search']['fit'][isub]=merit['fit']
        Res_data_vis['corr_search']['sig_corr'][isub]=p_best['sig_corr'].value    
        Res_data_vis['corr_search']['sig_uncorr'][isub]=p_best['sig_uncorr'].value 
    
    return None                



def binned_stddev_fit(param,x):
    """**Noise model.**
    
    Calculates white + red noise model as a function of spectral bin size.

    Args:
        TBD
    
    Returns:
        TBD
    
    """  
    #Uncorrelated standard deviation
    sig_uncorr=(param['sig_uncorr'].value)/np.sqrt(x)

    #Correlated standard deviation
    sig_corr=(param['sig_corr'].value)

    #Global standard deviation with bin size
    sig_bin=1./np.power( np.power(1./sig_uncorr,4.) + np.power(1./sig_corr,4.) ,1./4.)

    return sig_bin



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from scipy.interpolate import griddata
from copy import deepcopy
import bindensity as bind
import glob
from ..ANTARESS_conversions.ANTARESS_binning import calc_bin_prof,weights_bin_prof,init_bin_prof
from ..ANTARESS_grids.ANTARESS_prof_grid import init_custom_DI_prof,theo_intr2loc
from ..ANTARESS_grids.ANTARESS_occ_grid import init_surf_shift,def_surf_shift,sub_calc_plocc_spot_prop,up_plocc_prop
from ..ANTARESS_grids.ANTARESS_coord import excl_plrange
from ..ANTARESS_process.ANTARESS_data_align import align_data
from ..ANTARESS_analysis.ANTARESS_inst_resp import def_st_prof_tab,cond_conv_st_prof_tab,get_FWHM_inst,resamp_st_prof_tab,conv_st_prof_tab
from ..ANTARESS_general.utils import stop,dataload_npz,datasave_npz,closest_arr,np_where1D,gen_specdopshift,check_data

#%% Planet-occulted exposure profiles.

def def_in_plocc_profiles(inst,vis,gen_dic,data_dic,data_prop,coord_dic,system_param,theo_dic,glob_fit_dic,plot_dic):
    r"""**Planet-occulted exposure profiles.**
    
    Calls requested function to define planet-occulted profiles associated with each observed exposure
    
     - local profiles are used to correct residual profiles from stellar contamination
     - intrinsic profiles are used to assess fit quality

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    print('   > Building estimates for planet-occulted stellar profiles')
    opt_dic = data_dic['Intr']['opt_loc_data_corr'] 
    corr_mode = opt_dic['corr_mode']
    text_print={
        'DIbin':    '         Using DI master',
        'Intrbin':  '         Using binned intrinsic profiles',
        'glob_mod': '         Using global model',
        'indiv_mod':'         Using individual models',
        'rec_prof': '         Using reconstruction',
        }[corr_mode]
    print(text_print)  

    #Calculating
    if (gen_dic['calc_loc_data_corr']):
        print('         Calculating data')     
 
        #Using disk-integrated master or binned intrinsic profiles
        if corr_mode in ['DIbin','Intrbin']:
            data_add = plocc_prof_meas(opt_dic,corr_mode,inst,vis,data_dic['Intr'],gen_dic,data_dic,data_prop,data_dic['Atm'],coord_dic)
            
        #Using global profile model
        elif corr_mode=='glob_mod': 
            data_add = plocc_spocc_prof_globmod(opt_dic,corr_mode,inst,vis,gen_dic,data_dic,data_prop,system_param,theo_dic,coord_dic,glob_fit_dic,False)
            
        #Using individual profile models
        elif corr_mode=='indiv_mod': 
            data_add = plocc_prof_indivmod(opt_dic,corr_mode,inst,vis,gen_dic,data_dic)

        #Defining undefined pixels via a polynomial fit to defined pixels in complementary exposures, or via a 2D interpolation over complementary exposures and a narrow spectral band
        elif corr_mode=='rec_prof': 
            data_add = plocc_prof_rec(opt_dic,corr_mode,inst,vis,gen_dic,data_dic,coord_dic)
            
        #Saving complementary data for plots
        if plot_dic['map_Intr_prof_res']!='':
            data_add['loc_data_corr_inpath'] = data_dic[inst][vis]['proc_Intr_data_paths']
            data_add['loc_data_corr_outpath'] = data_dic[inst][vis]['proc_Res_data_paths']
            data_add['rest_frame'] = data_dic['Intr'][inst][vis]['rest_frame']
        datasave_npz(gen_dic['save_data_dir']+'Loc_estimates/'+corr_mode+'/'+inst+'_'+vis+'_add',data_add)
            
    #Checking that local data has been calculated for all exposures
    else:
        idx_est_loc = dataload_npz(gen_dic['save_data_dir']+'Loc_estimates/'+corr_mode+'/'+inst+'_'+vis+'_add')['idx_est_loc']
        data_paths={i_in:gen_dic['save_data_dir']+'Loc_estimates/'+corr_mode+'/'+inst+'_'+vis+'_'+str(i_in) for i_in in idx_est_loc}
        check_data(data_paths)

    return None



def plocc_prof_meas(opt_dic,corr_mode,inst,vis,gen_dic,data_dic,data_prop,coord_dic):
    r"""**Planet-occulted exposure profiles: measured**
    
    Sub-function to define planet-occulted profiles using measured profiles as best estimates of intrinsic profiles
    
    When binned intrinsic profiles are used
    
     - we first define the range and center of the binned profiles, along the bin dimension
     - for each in-transit local profile, we find the nearest binned profile along the bin dimension (via their centers)
     - we use intrinsic profiles that have already been aligned to a null rest frame, first shifting them to the rv of the target local profile, and then binning them
     - we could use the original intrinsic profiles and shift them by their own rv minus that of the local profile, however in this way we can use directly the outputs of `align_Intr()`, and 
       shifting the profiles in two steps is not an issue when they are not resampled until binned
     - we finally bin the aligned profiles, and scale the binned profile to the level of the local one 

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    data_inst = data_dic[inst]
    data_vis = data_inst[vis]
    
    #Initialize surface RV
    ref_pl,dic_rv,idx_aligned = init_surf_shift(gen_dic,inst,vis,data_dic,data_dic['Intr']['align_mode'])  

    #Binning mode
    #   - using current visit exposures only, or exposures from multiple visits
    if (inst in opt_dic['vis_in_bin']) and (len(opt_dic['vis_in_bin'][inst])>0):vis_to_bin = opt_dic['vis_in_bin'][inst]   
    else:vis_to_bin = [vis]           
    
    #Initialize binning
    #    - when disk-integrated profiles are used, the output tables contain a single value, associated with the single master (=binned profile) used for the extraction
    if corr_mode=='DIbin':
        in_type='DI'
        dim_bin = 'phase'
    elif corr_mode=='Intrbin':
        in_type='Intr'
        dim_bin = opt_dic['dim_bin']
        if not gen_dic['align_Intr']:stop('Intrinsic profiles must have been aligned')
    new_x_cen,_,_,x_cen_all,n_in_bin_all,idx_to_bin_all,dx_ov_all,_,idx_bin2orig,idx_bin2vis,idx_to_bin_unik = init_bin_prof(in_type,opt_dic[inst][vis],opt_dic['idx_in_bin'],dim_bin,coord_dic,inst,vis_to_bin,data_dic,gen_dic)

    #Find binned profile closest (along bin dimension ) to each processed in-transit exposure 
    if corr_mode=='Intrbin':idx_bin_closest = closest_arr(new_x_cen, x_cen_all[idx_aligned])
    
    #Processing in-transit exposures for which planet-occulted rv is known
    for isub,i_in in enumerate(idx_aligned):    
    
        #Upload spectral tables from residual or intrinsic profile of current exposure
        if data_dic['Intr']['plocc_prof_type']=='Intr':iexp_eff = i_in
        elif data_dic['Intr']['plocc_prof_type']=='Res':iexp_eff = gen_dic[inst][vis]['idx_in2exp'][i_in]
        data_loc_exp = dataload_npz(data_vis['proc_'+data_dic['Intr']['plocc_prof_type']+'_data_paths']+str(iexp_eff))

        #Index of binned profile associated with current processed exposure
        if corr_mode=='DIbin':i_bin = 0
        elif corr_mode=='Intrbin':i_bin = idx_bin_closest[isub]
                
        #Calculating binned profile associated with current processed exposure
        #    - since the shift is specific to each processed exposure, contributing profiles must be aligned and resampled for each one, either on the table common to all profiles, or on the table of current exposure, before being binned together        
        data_to_bin={}
        for iexp_off in idx_to_bin_all[i_bin]:    
        
            #Original index and visit of contributing exposure
            iexp_bin = idx_bin2orig[iexp_off]
            vis_bin = idx_bin2vis[iexp_off]
            
            #Upload latest processed disk-integrated data         
            if corr_mode=='DIbin':
                iexp_bin_glob = iexp_bin
                data_exp_bin = np.load(data_inst[vis_bin]['proc_DI_data_paths']+str(iexp_bin)+'.npz',allow_pickle=True)['data'].item() 
                if gen_dic['flux_sc']: scaled_data_paths = data_dic[inst][vis_bin]['scaled_DI_data_paths']  
                else:scaled_data_paths=None
                
                #Exclude planet-contaminated bins 
                #    - here we set to nan the flux (rather than just the defined bins) when profiles are still aligned in the star rest frame, to avoid having to shift for every exposure the planet-excluded ranges
                #      after profiles are aligned to the local surface velocity, the defined pixels will account for the exclusion and be used to define the weights in the bin routine
                if ('DI_Mast' in data_dic['Atm']['no_plrange']) and (iexp_bin in data_dic['Atm'][inst][vis_bin]['iexp_no_plrange']):
                    for iord in range(data_inst['nord']):
                        data_exp_bin['flux'][iord][excl_plrange(data_exp_bin['cond_def'][iord],data_dic['Atm'][inst][vis_bin]['exclu_range_star'],iexp_bin,data_loc_exp['edge_bins'][iord],data_vis['type'])[0]] = np.nan
                if data_dic[inst][vis_bin]['tell_sp']:data_exp_bin['tell'] = dataload_npz(data_dic[inst][vis_bin]['tell_DI_data_paths'][iexp_bin])['tell']             
                if data_dic[inst][vis_bin]['mean_gdet']:data_exp_bin['mean_gdet'] = dataload_npz(data_dic[inst][vis_bin]['mean_gdet_DI_data_paths'][iexp_bin])['mean_gdet']             
                data_ref = dataload_npz(data_dic[inst][vis_bin]['mast_DI_data_paths'][iexp_bin])
                
                
            #Upload intrinsic stellar profiles aligned in their local frame
            #    - we use intrinsic profiles already aligned in a null rest frame (still defined on their original tables unless a common table was used)     
            #    - profiles could be retrieved from their original extraction rest frame, and shifted directly to the surface rv associated with current exposure
            #      however in the ideal processing profiles were maintained on their individual tables, which were shifted without resampling to create the aligned profiles, thus doing an additional shift here
            # will not create any correlations 
            elif corr_mode=='Intrbin':
                iexp_bin_glob = gen_dic[inst][vis_bin]['idx_in2exp'][iexp_bin]
                if iexp_bin_glob not in data_dic['Intr'][inst][vis]['idx_def']:stop('Intrinsic exposure at i=',str(iexp_bin),' has not been aligned')
                data_exp_bin = dataload_npz(gen_dic['save_data_dir']+'Aligned_Intr_data/'+inst+'_'+vis_bin+'_'+str(iexp_bin))
                if data_dic[inst][vis_bin]['tell_sp']:data_exp_bin['tell'] = dataload_npz(gen_dic['save_data_dir']+'Aligned_Intr_data/'+inst+'_'+vis_bin+'_in_tell'+str(iexp_bin))['tell']             
                if data_dic[inst][vis_bin]['mean_gdet']:data_exp_bin['mean_gdet'] = dataload_npz(gen_dic['save_data_dir']+'Aligned_Intr_data/'+inst+'_'+vis_bin+'_in_mean_gdet'+str(iexp_bin))['mean_gdet']             
                data_ref = np.load(gen_dic['save_data_dir']+'Aligned_Intr_data/'+inst+'_'+vis_bin+'_in_ref'+str(iexp_bin)+'.npz',allow_pickle=True)['data'].item() 
                if gen_dic['flux_sc']: scaled_data_paths = data_dic[inst][vis_bin]['scaled_Intr_data_paths']  
                else:scaled_data_paths=None
                
            #Radial velocity shifts set to the opposite of the planet-occulted surface rv associated with current exposure
            rv_surf_star,rv_surf_star_edge = def_surf_shift(data_dic['Intr']['align_mode'],dic_rv,i_in,data_exp_bin,ref_pl,data_vis['type'],data_dic[inst]['system_prop'],data_dic[inst][vis_bin]['dim_exp'],data_dic[inst]['nord'],data_dic[inst][vis_bin]['nspec'])
            rv_shift_cen = -rv_surf_star
            spec_dopshift_cen = gen_specdopshift(rv_surf_star)
            if rv_surf_star_edge is not None:
                rv_shift_edge = -rv_surf_star_edge
                spec_dopshift_edge = gen_specdopshift(rv_surf_star_edge)
            else:
                rv_shift_edge = None
                spec_dopshift_edge = None            
                    
            #Aligning contributing profile at current exposure stellar surface rv 
            #    - aligned profiles are here resampled on the table of current exposure, which is common to all exposures if a common table is used        
            #    - complementary tables follow the same shifts
            data_to_bin[iexp_off]=align_data(data_exp_bin,data_vis['type'],data_dic[inst]['nord'],data_dic[inst][vis]['dim_exp'],gen_dic['resamp_mode'],data_loc_exp['cen_bins'],data_loc_exp['edge_bins'],rv_shift_cen,spec_dopshift_cen,spec_dopshift_edge = spec_dopshift_edge,rv_shift_edge = rv_shift_edge)
                
            #Shifting weighing master at current exposure stellar surface rv 
            #    - master will be resampled on the same table as current exposure
            data_ref_align=align_data(data_ref,data_vis['type'],data_dic[inst]['nord'],data_dic[inst][vis]['dim_exp'],gen_dic['resamp_mode'],data_loc_exp['cen_bins'], data_loc_exp['edge_bins'],rv_shift_cen,spec_dopshift_cen,spec_dopshift_edge = spec_dopshift_edge,rv_shift_edge = rv_shift_edge)

            #Weight profile
            data_to_bin[iexp_off]['weight'] = weights_bin_prof(range(data_inst['nord']),scaled_data_paths,inst,vis_bin,gen_dic['corr_Fbal'],gen_dic['corr_FbalOrd'],gen_dic['save_data_dir'],gen_dic['type'],data_inst['nord'],iexp_bin_glob,in_type,data_vis['type'],data_vis['dim_exp'],data_to_bin[iexp_off]['tell'],data_to_bin[iexp_off]['mean_gdet'],data_to_bin[iexp_off]['cen_bins'],coord_dic[inst][vis_bin]['t_dur'][iexp_off],data_ref_align['flux'],data_ref_align['cov'],bdband_flux_sc = gen_dic['flux_sc'])            
            
        #Calculating binned profile
        data_est_loc = calc_bin_prof(idx_to_bin_all[i_bin],data_dic[inst]['nord'],data_vis['dim_exp'],data_vis['nspec'],data_to_bin,inst,n_in_bin_all[i_bin],data_loc_exp['cen_bins'],data_loc_exp['edge_bins'],dx_ov_in = dx_ov_all[i_bin])
        
        #Rescaling measured intrinsic profile to the level of the local profile
        #    - this operation assumes that all exposures used to compute the master-out have not been rescaled with respect to the reference, ie that they all have the same flux balance as current exposure before it was rescaled
        #    - see rescale_profiles() and proc_intr_data() for more details
        #      a given local profile write as 
        #      F_res(w,t,vis) = MFstar(w,vis) - Fsc(w,vis,t)
        #   at low resolution, in the continuum (see rescale_profiles())
        #      Fsc(w,vis,t) = LC(w,t)*Fstar(w,vis_norm)*Cref(w)
        #   assuming that the rescaling light curves are constant outside of the transit, all out-of-transit profiles are equivalent between themselves and thus to the master-out
        #      MFstar(w,vis) ~ Fstar(w,vis_norm)*Cref(w)
        #   thus
        #      F_res(w,t,vis) = Fstar(w,vis_norm)*Cref(w) - LC(w,t)*Fstar(w,vis_norm)*Cref(w)
        #                     = Fstar(w,vis_norm)*Cref(w)*(1 - LC(w,t))
        #                     = MFstar(w,vis)*(1 - LC(w,t))           
        #    - if intrinsic profiles are requested no scaling is applied, since F_res(w,t,vis) = F_intr(w,t,vis)*(1 - LC(w,t))      
        #    - the scaling spectrum is defined at all pixels, and thus does not affect undefined pixels in the master (the covariance matrix cannot be sliced)
        if (data_dic['Intr']['plocc_prof_type']=='Res') and gen_dic['flux_sc']:
            data_scaling = dataload_npz(data_vis['scaled_Intr_data_paths']+str(gen_dic[inst][vis]['idx_in2exp'][i_in]))
            for iord in range(data_inst['nord']):
                loc_flux_scaling_ord = data_scaling['loc_flux_scaling'](data_est_loc['cen_bins'][iord])              
                data_est_loc['flux'][iord],data_est_loc['cov'][iord] = bind.mul_array(data_est_loc['flux'][iord],data_est_loc['cov'][iord],loc_flux_scaling_ord)
            
        #Saving estimate of local profile for current exposure                   
        np.savez_compressed(gen_dic['save_data_dir']+'Loc_estimates/'+corr_mode+'/'+inst+'_'+vis+'_'+str(i_in),data=data_est_loc,allow_pickle=True)

    #Complementary data
    data_add={'idx_est_loc':idx_aligned}
    if corr_mode=='Intrbin':data_add['idx_bin_closest']=idx_bin_closest

    return data_add





def plocc_spocc_prof_globmod(opt_dic,corr_mode,inst,vis,gen_dic,data_dic,data_prop,system_param,theo_dic,coord_dic,glob_fit_dic,spot_on):
    r"""**Planet-occulted / spotted exposure profiles: global model**
    
    Sub-function to define planet-occulted and spotted profiles using line profile models fitted to all
        - intrinsic profiles together in `fit_IntrProf_all()`.
          This model can be based on analytical, measured, or theoretical profiles.        
        - to all differential profiles together with `fit_ResProf_all()`.
          This model is based on analytical profiles.  
    
    Flux scaling is not applied to a global intrinsic profile using the chromatic light curve, but re-calculated for each planet-occulted intrinsic line profile for more flexibility (the cumulated scaling should be equivalent).
    rv used to shift the profiles are similarly re-calculated theoretically. 

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    data_vis = data_dic[inst][vis]
    def_iord = opt_dic['def_iord']
    if (data_dic['Intr']['align_mode']=='meas'):stop('This mode cannot be used with measured RVs')

    #Retrieving selected model properties
    fit_txt = None
    if (not spot_on) and ('IntrProf_prop_path' not in opt_dic):fit_txt = 'Intr'
    if (spot_on) and ('ResProf_prop_path' not in opt_dic):fit_txt = 'Res'
    if fit_txt is not None:
        fit_subdirs = glob.glob(gen_dic['save_data_dir']+'/Joined_fits/'+fit_txt+'Prof/*/')
        if len(fit_subdirs)==0:stop('         No existing fit directory. Run a fit to the '+{'Intr':'intrinsic','Res':'differential'}[fit_txt]+' line profiles.')
        else:
            fit_types = [fit_subdir.split('/')[-2] for fit_subdir in fit_subdirs]
            if 'mcmc' in fit_types:fit_used  ='mcmc'
            else:fit_used  ='chi2'
            print('         Using existing '+fit_used+' fit as default')
            opt_dic[fit_txt+'Prof_prop_path']={inst:{vis:gen_dic['save_data_dir']+'/Joined_fits/'+fit_txt+'Prof/'+fit_used+'/Fit_results'}}  
    if spot_on:prof_type = 'Res'
    else:prof_type='Intr'        
    data_prop = dataload_npz(opt_dic[prof_type+'Prof_prop_path'][inst][vis])
    params=data_prop['p_final'] 
    fixed_args={  
        'mode':opt_dic['mode'],
        'type':data_vis['type'],
        'nord':data_dic[inst]['nord'],
        'nthreads': opt_dic['nthreads'],
        'resamp_mode' : gen_dic['resamp_mode'], 
        'inst':inst,
        'vis':vis,  
        'fit':False,
        'ph_fit':data_prop['ph_fit'],
        'system_param':system_param,
        'genpar_instvis':data_prop['genpar_instvis'],
        'name_prop2input':data_prop['name_prop2input'],
        'fit_orbit':data_prop['fit_orbit'],
        'fit_RpRs':data_prop['fit_RpRs'],
        'var_par_list':data_prop['var_par_list'],
        'system_prop':data_prop['system_prop'],
        'grid_dic':data_prop['grid_dic'],
    } 
    if fixed_args['mode']=='ana':
        fixed_args.update({  
            'mac_mode':theo_dic['mac_mode'], 
            'coeff_line':data_prop['coeff_line_dic'][inst][vis],
            'func_prof_name':data_prop['func_prof_name'][inst]
        })        
        for key in ['coeff_ord2name','pol_mode','coord_line','linevar_par']:fixed_args[key] = data_prop[key]
    if spot_on:  
        fixed_args.update({          
        'fit_spot_ang':data_prop['fit_spot_ang'],
        'fit_spot':data_prop['fit_spot'],
        'system_spot_prop':data_prop['system_spot_prop'],
        'update_crosstime':False,    
        'conv2intr' :False,           
            })
        transit_spots=data_vis['transit_sp']
        iexp_list = range(data_vis['n_in_visit'])
        spots_prop = data_vis['spots_prop']
        plocc_prof_type = 'Res'
        
        #Defining parameters for the clean version of profiles
        if opt_dic['clean_calc']:
            clean_params = deepcopy(params)
            clean_params['use_spots']=False        
        
    else:
        fixed_args.update({          
        'system_spot_prop':{},          
            })
        plocc_prof_type = data_dic['Intr']['plocc_prof_type']
        if data_dic['Intr']['plocc_prof_type']=='Intr':fixed_args['conv2intr'] = True
        else:fixed_args['conv2intr'] = False 
        transit_spots={}
        spots_prop ={}
        iexp_list = range(data_vis['n_in_tr'])
    chrom_mode = data_vis['system_prop']['chrom_mode']
  
    #Activation of spectral conversion and resampling 
    cond_conv_st_prof_tab(theo_dic['rv_osamp_line_mod'],fixed_args,data_vis['type']) 

    #Updating coordinates with the best-fit properties
    ph_rec = {}
    coord_vis = coord_dic[inst][vis]
    for pl_loc in data_vis['transit_pl']:
        ph_rec[pl_loc] = np.vstack((coord_vis[pl_loc]['st_ph'],coord_vis[pl_loc]['cen_ph'],coord_vis[pl_loc]['end_ph']) ) 
    _,coord_pl_sp,_ = up_plocc_prop(inst,vis,fixed_args,params,data_vis['transit_pl'],ph_rec,coord_vis,transit_spots=transit_spots)

    #Processing relevant exposures
    for isub,iexp in enumerate(iexp_list):
        if not spot_on:
            iexp_glob = gen_dic[inst][vis]['idx_in2exp'][iexp]
            if plocc_prof_type=='Intr':iexp_eff = iexp
            else:iexp_eff = iexp_glob
        else:
            iexp_glob =iexp
            iexp_eff = iexp

        #Upload spectral tables from residual profile of current exposure
        data_loc_exp = dataload_npz(data_vis['proc_'+plocc_prof_type+'_data_paths']+str(iexp_eff))
    
        #Limit model table to requested definition range
        if len(opt_dic['def_range'])==0:cond_calc_pix = np.ones(data_vis['nspec'] ,dtype=bool)    
        else:cond_calc_pix = (data_loc_exp['edge_bins'][def_iord][0:-1]>=opt_dic['def_range'][0]) & (data_loc_exp['edge_bins'][def_iord][1:]<=opt_dic['def_range'][1])             
        idx_calc_pix = np_where1D(cond_calc_pix)
        cond_def_full = np.zeros(data_vis['nspec'],dtype=bool)
        cond_def_full[idx_calc_pix] = True
    
        #Saves
        data_store = {'cen_bins':data_loc_exp['cen_bins'],'edge_bins':data_loc_exp['edge_bins'],'cond_def':np.array([cond_def_full])}
    
        #Final table for model line profile
        fixed_args['ncen_bins']=len(idx_calc_pix)
        fixed_args['dim_exp'] = [1,fixed_args['ncen_bins']] 
        fixed_args['cen_bins'] = data_loc_exp['cen_bins'][def_iord,idx_calc_pix]
        fixed_args['edge_bins']=data_loc_exp['edge_bins'][def_iord,idx_calc_pix[0]:idx_calc_pix[-1]+2]
        fixed_args['dcen_bins']=fixed_args['edge_bins'][1::] - fixed_args['edge_bins'][0:-1] 
    
        #Initializing stellar profiles
        #    - can be defined using the first exposure table
        if isub==0:
            fixed_args = init_custom_DI_prof(fixed_args,gen_dic,data_vis['system_prop'],spots_prop,theo_dic,system_param['star'],params)                  
    
            #Effective instrumental convolution
            fixed_args['FWHM_inst'] = get_FWHM_inst(inst,fixed_args,fixed_args['cen_bins'])
    
        #Resampled spectral table for model line profile
        if fixed_args['resamp']:resamp_st_prof_tab(None,None,None,fixed_args,gen_dic,1,theo_dic['rv_osamp_line_mod'])
        
        #Table for model calculation
        args_exp = def_st_prof_tab(None,None,None,fixed_args) 
    
        #Define broadband scaling of intrinsic profiles into local profiles
        if plocc_prof_type=='Res':
            args_exp['Fsurf_grid_spec'] = theo_intr2loc(fixed_args['grid_dic'],fixed_args['system_prop'],args_exp,args_exp['ncen_bins'],fixed_args['grid_dic']['nsub_star']) 
    
        #Planet-occulted properties
        surf_prop_dic,spot_prop_dic,_ = sub_calc_plocc_spot_prop([chrom_mode],args_exp,['line_prof'],data_vis['transit_pl'],deepcopy(system_param),theo_dic,fixed_args['system_prop'],params,coord_pl_sp,[iexp_glob],system_spot_prop_in=fixed_args['system_spot_prop'])

        #With spots
        if spot_on:
            save_path = gen_dic['save_data_dir']+'Diff_estimates/'+corr_mode+'/'+inst+'_'+vis+'_'+str(iexp)

            #Planet-occulted and spotted line profiles - unclean
            pl_line_model = surf_prop_dic[chrom_mode]['line_prof'][:,0]
            sp_line_model = spot_prop_dic[chrom_mode]['line_prof'][:,0]
            line_prof_cons = {'unclean_pl_flux':pl_line_model, 'unclean_sp_flux':sp_line_model}
            
            #Planet-occulted and spotted line profiles - clean
            if opt_dic['clean_calc']:
                
                #Only planet-occulted profiles
                clean_surf_prop_dic,_,_ = sub_calc_plocc_spot_prop([chrom_mode],args_exp,['line_prof'],data_vis['transit_pl'],deepcopy(system_param),theo_dic,fixed_args['system_prop'],clean_params,coord_pl_sp,[iexp])
                
                #Only spotted profiles
                _,clean_spot_prop_dic,_ = sub_calc_plocc_spot_prop([chrom_mode],args_exp,['line_prof'],[],deepcopy(system_param),theo_dic,fixed_args['system_prop'],params,coord_pl_sp,[iexp],system_spot_prop_in=fixed_args['system_spot_prop'])
    
                #Storing
                clean_pl_line_model = clean_surf_prop_dic[chrom_mode]['line_prof'][:,0]
                clean_sp_line_model = clean_spot_prop_dic[chrom_mode]['line_prof'][:,0]
                line_prof_cons.update({'clean_pl_flux':clean_pl_line_model, 'clean_sp_flux':clean_sp_line_model})

        #Without spots
        else:
            save_path = gen_dic['save_data_dir']+'Loc_estimates/'+corr_mode+'/'+inst+'_'+vis+'_'+str(iexp)
    
            #Planet-occulted line profile 
            sp_line_model = surf_prop_dic[chrom_mode]['line_prof'][:,0]
            line_prof_cons = {'flux':sp_line_model}
        
        #Exposure profiles
        for line_prof in list(line_prof_cons.keys()):
           
            #Scaling to fitted intrinsic continuum level
            #    - model profiles have been output with a continuum level unity (through the option 'conv2intr') and are thus scaled to the fitted level
            if plocc_prof_type=='Intr':line_prof_cons['flux']*=params['cont']
         
            #Conversion and resampling 
            flux_loc = conv_st_prof_tab(None,None,None,fixed_args,args_exp,line_prof_cons['flux'],fixed_args['FWHM_inst'])
        
            #Filling full table with defined reconstructed profile
            plocc_prof = np.zeros(data_vis['nspec'],dtype=float)*np.nan
            plocc_prof[idx_calc_pix] = flux_loc
            
            #Store
            data_store.update({line_prof:np.array([plocc_prof])})
        
        #Saving estimate of local profile for current exposure                 
        np.savez_compressed(save_path,data=data_store,allow_pickle=True)

    #Complementary data
    data_add={'idx_est_loc':iexp_list,'cont':params['cont']}

    return data_add



def plocc_prof_indivmod(opt_dic,corr_mode,inst,vis,gen_dic,data_dic):
    r"""**Planet-occulted exposure profiles: individual model**
    
    Sub-function to define planet-occulted profiles using line profile model fitted to individual intrinsic lines.
    These models correspond directly to the measured profile and needs only be rescaled to the level of the local profile.
    This approach only works for exposures in which the stellar line could be fitted correctly after excluding the planet-contaminated range

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    data_vis = data_dic[inst][vis]
    if data_vis['type']!='CCF':stop('Method not valid for spectra')

    #Upload fit results
    data_prop=(np.load(gen_dic['save_data_dir']+'Introrig_prop/'+inst+'_'+vis+'.npz',allow_pickle=True)['data'].item())['prof_fit_dic']
    idx_aligned = np_where1D(data_prop['cond_detected'])

    #Processing in-transit exposures
    for i_in in idx_aligned:
     
        #Model intrinsic profile
        data_est_loc={'cen_bins':data_prop[i_in]['cen_bins'],'edge_bins':data_prop[i_in]['edge_bins'],'flux':data_prop[i_in]['flux'][:,None],'cond_def':np.ones(data_vis['dim_exp'],dtype=bool)}

        #Rescaling model intrinsic profile to the level of the local profile
        if data_dic['Intr']['plocc_prof_type']=='Res':
            loc_flux_scaling = dataload_npz(data_vis['scaled_Intr_data_paths']+str(gen_dic[inst][vis]['idx_in2exp'][i_in]))['loc_flux_scaling']
            data_est_loc['flux'][0] *= loc_flux_scaling(data_est_loc['cen_bins'][0]) 

        #Saving estimate of local profile for current exposure                   
        np.savez_compressed(gen_dic['save_data_dir']+'Loc_estimates/'+corr_mode+'/'+inst+'_'+vis+'_'+str(i_in),data=data_est_loc,allow_pickle=True)

    #Complementary data
    data_add={'idx_est_loc':idx_aligned}
      
    return data_add



def plocc_prof_rec(opt_dic,corr_mode,inst,vis,gen_dic,data_dic,coord_dic):
    r"""**Planet-occulted exposure profiles: reconstructed**
    
    Sub-function to define planet-occulted profiles by reconstructing undefined pixels via a polynomial fit to defined pixels in complementary exposures, or via a 2D interpolation over complementary exposures and a narrow spectral band.
    
     - providing the planetary track shifts sufficiently during the transit, we can reconstruct the local spectra at all wavelengths by interpolating the surrounding spectra at the wavelengths masked for planetary absorption.
       this approach allows accounting for changes in the shape of the local stellar spectra between exposures.
     - we reconstruct undefined pixels in the map of intrinsic profiles aligned in the null rest frame and resampled on the common spectral table.
       for each exposure, we then shift the corresponding reconstructed profile to the stellar surface velocity associated with the exposure, and resample it on the exposure table.
       ideally the map should be first aligned at the stellar surface velocity of each exposure, and then resampled on its table, but that would take too much time and the error is likely lower than that introduced by the interpolation.
     - no errors are propagated onto the new pixels.
     - at defined pixels, the local profile and its best estimate for the intrinsic profile will have the same values, resulting in null values in the atmospheric profiles.
       there will be small differences due to the need to resample the estimate on a different table than the corresponding local profile.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    data_inst = data_dic[inst]
    data_vis = data_inst[vis]
    if not gen_dic['align_Intr']:stop('Intrinsic stellar profiles must have been aligned')

    #Upload common spectral table for processed visit
    data_com = np.load(data_inst[vis]['proc_com_data_paths']+'.npz',allow_pickle=True)['data'].item()  

    #Using current visit exposures only, or exposures from multiple visits
    if (inst in opt_dic['vis_in_rec']) and (len(opt_dic['vis_in_rec'][inst])>0):vis_in_rec = opt_dic['vis_in_rec'][inst]     
    else:vis_in_rec = [vis]  

    #Initializing reconstruction
    y_cen_all = np.zeros(0,dtype=float)
    idx_orig_vis = np.zeros(0,dtype=int)
    vis_orig=np.zeros(0,dtype='U35')
    n_in_rec = 0
    for vis_rec in vis_in_rec:

        #Original indexes of aligned intrinsic profiles in current visit, which can potentially contribute to the reconstruction
        #    - relative to in-transit tables
        ref_pl,dic_rv_rec,_ = init_surf_shift(gen_dic,inst,vis_rec,data_dic,data_dic['Intr']['align_mode'])
            
        #Limiting aligned intrinsic profiles to input selection
        idx_in_rec = data_dic['Intr'][inst][vis]['idx_def']
        if (inst in opt_dic['idx_in_rec']) and (vis_rec in opt_dic['idx_in_rec'][inst]):
            idx_in_rec = np.intersect1d(opt_dic['idx_in_rec'][inst][vis_rec],)
            if len(idx_in_rec)==0:stop('No remaining exposures after input selection')  
        n_to_rec = len(idx_in_rec)                   

        #Tables along the chosen fit/interpolation dimension
        if opt_dic['dim_bin']=='phase':    
            y_cen_vis_rec = coord_dic[inst][vis_rec]['cen_ph'][gen_dic[inst][vis_rec]['idx_in']]   
        elif opt_dic['dim_bin'] in ['xp_abs','r_proj']: 
            transit_prop_nom = (np.load(gen_dic['save_data_dir']+'Introrig_prop/PlOcc_Prop_'+inst+'_'+vis_rec+'.npz',allow_pickle=True)['data'].item())['achrom'][ref_pl]                           
            y_cen_vis_rec = transit_prop_nom[opt_dic['dim_bin']][0,:]
        y_cen_all=np.append(y_cen_all,y_cen_vis_rec[idx_in_rec])
        
        #Store properties for visit to reconstruct
        if vis_rec==vis:
            dic_rv=deepcopy(dic_rv_rec)
            idx_aligned=deepcopy(idx_in_rec)
            
            #Indexes of processed visit within data_for_rec tables
            isub_vis_to_rec = n_in_rec+np.arange(n_to_rec)

            #Coordinates of exposures in processed visit
            y_cen_vis = y_cen_vis_rec[idx_in_rec]        

        #Store exposure identifiers
        n_in_rec+=n_to_rec                                          #total number of contributing exposures
        idx_orig_vis = np.append(idx_orig_vis,idx_in_rec)                   #original indexes of contributing exposures in their respective in-transit tables
        vis_orig = np.append(vis_orig,np.repeat(vis_rec,len(idx_in_rec)))   #original visits of contributing exposures

    #Upload contributing and processed intrinsic profiles
    #    - data must be put in global tables to perform the fits and interpolations  
    #    - the fits cannot deal with covariance matrix along the dimension orthogonal to the spectral one, thus we only use the diagonal
    data_for_rec={}
    data_rec={}
    for key in ['flux','err']:data_for_rec[key]=np.zeros([n_in_rec]+data_vis['dim_exp'], dtype=float)*np.nan
    data_for_rec['cond_def']=np.zeros([n_in_rec]+data_vis['dim_exp'], dtype=bool)
    isub_to_rec=0
    for isub_rec,(vis_rec,iexp_rec) in enumerate(zip(vis_orig,idx_orig_vis)):
        data_exp = np.load(gen_dic['save_data_dir']+'Aligned_Intr_data/'+inst+'_'+vis_rec+'_'+str(iexp_rec)+'.npz',allow_pickle=True)['data'].item()    
    
        #Resampling aligned intrinsic profiles on the common spectral table of the processed visit
        #    - if the processed and reconstructed visits are the same and exposures do not share a common table
        #    - if the processed and reconstructed visits are not the same, and visits do not share a common table
        if ((vis_rec==vis) and (not data_vis['comm_sp_tab'])) or ((vis_rec!=vis) and (not data_inst['comm_sp_tab'])): 
            for iord in range(data_inst['nord']):
                data_for_rec['flux'][isub_rec,iord],cov_ord = bind.resampling(data_com['edge_bins'][iord],data_exp['edge_bins'][iord], data_exp['flux'][iord] , cov =  data_exp['cov'][iord], kind=gen_dic['resamp_mode']) 
                data_for_rec['err'][isub_rec,iord] = np.sqrt(cov_ord[0])
            data_for_rec['cond_def'][isub_rec] = ~np.isnan(data_for_rec['flux'][isub_rec])     

        #Aligned intrinsic profiles are already defined on a common table for all visits
        else:
            for key in ['flux','cond_def']:data_for_rec[key][isub_rec] = data_exp[key]
            for iord in range(data_inst['nord']):data_for_rec['err'][isub_rec,iord] = np.sqrt(data_exp['cov'][iord][0])
            
        #Initialize reconstructed profiles table for processed visit               
        if (vis_rec==vis):
            data_rec[isub_to_rec]={
                'cen_bins':data_com['cen_bins'],
                'edge_bins':data_com['edge_bins'],
                'flux':data_for_rec['flux'][isub_rec]}
            isub_to_rec+=1

    #Process each order 
    for iord in range(data_dic[inst]['nord']):
        
        #Identify pixels to reconstruct in at least one exposure of the processed visit
        idx_undef_pix = np_where1D(np.sum(data_for_rec['cond_def'][isub_vis_to_rec,iord,:],axis=0)<n_to_rec) 

        #Process undefined pixels
        for idx_pix in idx_undef_pix: 

            #Exposures for which the pixel is defined in contributing exposures from all visits
            cond_def_pix_exp = data_for_rec['cond_def'][:,iord,idx_pix]

            #Exposures for which the pixel is undefined in exposures of processed visit
            idx_undef_pix_vis = np_where1D( data_for_rec['cond_def'][isub_vis_to_rec,iord,idx_pix] )
            
            #------------------------------------------------------------------------------------------------------------------------
                                         
            #Pixel-per-pixel interpolation
            #    - for each undefined pixel, we perform a polynomial fit along the chosen dimension over defined pixels
            if opt_dic['rec_mode']=='pix_polfit':

                #Fit polynomial along the chosen dimension    
                #    - only if more than two pixels are defined
                if np.sum(cond_def_pix_exp)>2:
                    func_def_pix = np.poly1d(np.polyfit(y_cen_all[cond_def_pix_exp], data_for_rec['flux'][cond_def_pix_exp,iord,idx_pix],opt_dic['pol_deg'],w=1./data_for_rec['err'][cond_def_pix_exp,iord,idx_pix]))

                    #Calculate polynomial in exposures of processed visit for which the pixel is undefined
                    val_def_pix = func_def_pix(y_cen_vis[idx_undef_pix_vis])
                    for isub_to_rec,val_def_pix_loc in zip(idx_undef_pix_vis,val_def_pix):
                        data_rec[isub_to_rec]['flux'][iord,idx_pix] = val_def_pix_loc   

            #------------------------------------------------------------------------------------------------------------------------
                        
            #2D band interpolation
            elif opt_dic['rec_mode']=='band_interp':  
                
                #Pixels to interpolate
                #    - we use the pixels defined over complementary exposures in a band surrounding the current pixel
                #    - tables must be put in 1D
                idx_band = np.arange(max(0,idx_pix-opt_dic['band_pix_hw']),min(data_vis['nspec']-1,idx_pix+opt_dic['band_pix_hw'])+1)
                nx_band = len(idx_band)
                cond_def_band = data_for_rec['cond_def'][:,iord,idx_band].flatten()      #defined pixels in 1D flux table within the band
                if np.sum(cond_def_band)>0:
                    flux_band = (data_for_rec['flux'][:,iord,idx_band].flatten())[cond_def_band]                       
                    x_cen_band = np.repeat(data_com['cen_bins'][iord,idx_band],n_in_rec)[cond_def_band]
                    y_cen_band = np.tile(y_cen_all,[nx_band])[cond_def_band]

                    #Tables must have the same scaling because the interpolation is done in euclidian norm
                    xmin_sc=np.min(x_cen_band)
                    ymin_sc=np.min(y_cen_band)
                    xscale=np.max(x_cen_band)-xmin_sc
                    yscale=np.max(y_cen_band)-ymin_sc
                    if (xscale>0) and (yscale>0):

                        #Pixels to reconstruct in processed visit
                        n_undef_pix_exp_vis = len(idx_undef_pix_vis)
                        x_cen_rec = np.repeat(data_com['cen_bins'][iord,idx_pix],n_undef_pix_exp_vis)
                        y_cen_rec = y_cen_vis[idx_undef_pix_vis]
                        
                        #Scaling
                        x_cen_band_sc=(x_cen_band-xmin_sc)/xscale
                        y_cen_band_sc=(y_cen_band-ymin_sc)/yscale    
                        x_cen_rec_sc=(x_cen_rec-xmin_sc)/xscale
                        y_cen_rec_sc=(y_cen_rec-ymin_sc)/yscale     
 
                        #Interpolating
                        val_def_pix = griddata((x_cen_band_sc,y_cen_band_sc), flux_band, (x_cen_rec_sc, y_cen_rec_sc),method=opt_dic['interp_mode'])
                        for isub_to_rec,val_def_pix_loc in zip(idx_undef_pix_vis,val_def_pix):
                            data_rec[isub_to_rec]['flux'][iord,idx_pix] = val_def_pix_loc

    #------------------------------------------------------------------------------------------------------------------------

    #Processing in-transit exposures with reconstructed intrinsic profiles
    for isub,i_in in enumerate(idx_aligned):

        #Upload residual or intrinsic profile for current exposure to get its spectral tables
        if data_dic['Intr']['plocc_prof_type']=='Intr':iexp_eff = i_in
        else:iexp_eff = gen_dic[inst][vis]['idx_in2exp'][i_in]
        data_loc_exp = dataload_npz(data_vis['proc_'+data_dic['Intr']['plocc_prof_type']+'_data_paths']+str(iexp_eff))  

        #Radial velocity shifts set to the opposite of the planet-occulted surface rv associated with current exposure
        rv_surf_star,rv_surf_star_edge = def_surf_shift(data_dic['Intr']['align_mode'],dic_rv,i_in,data_rec[isub],ref_pl,data_vis['type'],data_vis['system_prop'],data_vis['dim_exp'],data_dic[inst]['nord'],data_vis['nspec'])

        #Aligning reconstructed profile at current exposure stellar surface rv 
        #    - reconstructed profile is already aligned in a null rest frame and resampled on a common table
        #    - aligned profiles are resampled on the table of current exposure, which is common to all exposures if a common table is used 
        rv_shift_cen = -rv_surf_star
        spec_dopshift_cen = gen_specdopshift(rv_surf_star)
        if rv_surf_star_edge is not None:
            rv_shift_edge = -rv_surf_star_edge
            spec_dopshift_edge = gen_specdopshift(rv_surf_star_edge)
        else:
            rv_shift_edge = None
            spec_dopshift_edge = None  
        data_est_loc=align_data(data_rec[isub],data_vis['type'],data_dic[inst]['nord'],data_vis['dim_exp'],gen_dic['resamp_mode'],data_loc_exp['cen_bins'],data_loc_exp['edge_bins'],rv_shift_cen,spec_dopshift_cen,rv_shift_edge = rv_shift_edge,spec_dopshift_edge = spec_dopshift_edge, nocov = True)

        #Rescaling reconstructed intrinsic profile to the level of the local profile
        if data_dic['Intr']['plocc_prof_type']=='Res':
            loc_flux_scaling = dataload_npz(data_vis['scaled_Intr_data_paths']+str(gen_dic[inst][vis]['idx_in2exp'][i_in]))['loc_flux_scaling']
            for iord in range(data_dic[inst]['nord']):data_est_loc['flux'][iord] *=loc_flux_scaling(data_est_loc['cen_bins'][iord]) 

        #Saving estimate of local profile for current exposure                   
        np.savez_compressed(gen_dic['save_data_dir']+'Loc_estimates/'+corr_mode+'/'+inst+'_'+vis+'_'+str(i_in),data=data_est_loc,allow_pickle=True)


    #Complementary data
    data_add={'idx_est_loc':idx_aligned}

    return data_add



#%% Differential exposure profiles.
#    - accounting for planet-occulted and spotted stellar profiles

def def_diff_profiles(inst,vis,gen_dic,data_dic,data_prop,coord_dic,system_param,theo_dic,glob_fit_dic,plot_dic):
    r"""**Planet-occulted and spotted exposure profiles.**
    
    Calls requested function to define planet-occulted and spotted profiles associated with each observed exposure

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    print('   > Building estimates for planet-occulted and spotted stellar profiles')
    if data_dic[inst][vis]['spots_prop']=={}:stop('Spot properties are not provided. We recommended using the separate routine dedicated to the extraction of planet-occulted profiles instead.')
    opt_dic = data_dic['Res']['opt_loc_data_corr'] 
    corr_mode = opt_dic['corr_mode']
    print('         Using global model')  

    #Calculating
    if (gen_dic['calc_res_loc_data_corr']):
        print('         Calculating data')     
             
        #Calculating the clean version of the data
        opt_dic['clean_calc']=False
        if (plot_dic['map_Res_prof_clean_pl_est']!='') or (plot_dic['map_Res_prof_clean_sp_est']!=''):
            opt_dic['clean_calc']=True

        #Using global profile model
        if corr_mode=='glob_mod': 
            data_add = plocc_spocc_prof_globmod(opt_dic,corr_mode,inst,vis,gen_dic,data_dic,data_prop,system_param,theo_dic,coord_dic,glob_fit_dic,False)
        
        else:stop('WARNING: Only joined-fit results can be used at the moment. Set corr_mode to \'glob_mod\'')

        #Saving complementary data for plots
        if (plot_dic['map_Res_prof_clean_sp_res']!='') or (plot_dic['map_Res_prof_clean_pl_res']!='') or (plot_dic['map_Res_prof_unclean_sp_res']!='') or (plot_dic['map_Res_prof_unclean_pl_res']!=''):
            data_add['loc_data_corr_path'] = data_dic[inst][vis]['proc_Res_data_paths']
            data_add['rest_frame'] = data_dic['Res'][inst][vis]['rest_frame']
        datasave_npz(gen_dic['save_data_dir']+'Diff_estimates/'+corr_mode+'/'+inst+'_'+vis+'_add',data_add)

    #Checking that local data has been calculated for all exposures
    else:
        idx_est_loc = dataload_npz(gen_dic['save_data_dir']+'Diff_estimates/'+corr_mode+'/'+inst+'_'+vis+'_add')['idx_est_loc']
        data_paths={i_in:gen_dic['save_data_dir']+'Diff_estimates/'+corr_mode+'/'+inst+'_'+vis+'_'+str(i_in) for i_in in idx_est_loc}
        check_data(data_paths)

    return None










#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
import lmfit
from lmfit import Parameters
from scipy.interpolate import interp1d
from scipy import stats
from copy import deepcopy
from pathos.multiprocessing import Pool
from numpy.polynomial import Polynomial
from ..ANTARESS_conversions.ANTARESS_binning import sub_calc_bins,sub_def_bins
from ..ANTARESS_analysis.ANTARESS_inst_resp import return_resolv
from ..ANTARESS_general.utils import dataload_npz,np_where1D,stop,init_parallel_func,check_data
from ..ANTARESS_general.minim_routines import call_lmfit




def calc_gcal(gen_dic,data_dic,inst,plot_dic,coord_dic):
    r"""**Main instrumental calibration routine.**
    
    Estimates instrumental calibration (see `weights_bin_prof()`)
    
    
    Calibration is used as weight or to scale back profiles approximatively to their raw photoelectron counts during CCF calculations (this allows artifically increasing errors when combining spectral regions with different SNRs).
    Spectra must however be kept to their extracted flux units througout ANTARESS processing, as the stellar spectrum shifts over time in the instrument rest frame.
    The same region of the spectrum thus sees different instrumental responses over time and is measured with different flux levels
    shifted stellar spectra with the same profile would thus have a different color balance after being converted in raw count units, preventing in particular the correct calculation of binned spectra 
    
    Weights are used for temporal binning, and are relevant only if the weight of a given pixel change over time, the calibration is thus not defined for CCFs.
    
    A median calibration profile over all visits of an instrument is calculated in spectral mode. 
    We first fit a model to the calibration for each exposure, so that it can be extrapolated over a larger, common range for all visits 
    then we calculate the median of these extrapolated model, interpolate it as a function, and use it to define the common calibration profile over the specific table of each exposure.
    
    Overall the estimated calibrations are stable between exposures but low count levels ten to yield larger calibrations - hence the independent calculation of calibration per exposure, before taking the median over the visit.
    This may come from additional noise sources.
    If :math:`E = E_\mathrm{white} + E_\mathrm{red} = \sqrt{g_\mathrm{true}} \sqrt{F} + E_\mathrm{red}` then :math:`g_\mathrm{true} = (E-E_\mathrm{red})^2/F < E^2/F = g_\mathrm{meas}` (since :math:`E_\mathrm{red} < 2 E`), so that we underestimate the actual calibration. 
    In this case our assumptions to estimate gdet do not hold anymore, but for the purpose of weighing the exposures and scaling to raw count levels we still use the measured calibration.
    Since :math:`g_\mathrm{meas} = (E_\mathrm{white} + E_\mathrm{red})^2/F` this factor accounts for the true calibration but also for additional noise sources.        
    
    Spectra are typically provided in the solar barycentric rest frame, and are thus not defined here in the detector rest frame, so that the calibration profiles in different epochs may be shifted by the Earth barycentric RV difference.   
    We nonetheless use a single calibration profile, constant in time and common to all processed exposures of an instrument, so that the relative color balance between spectra is not modified when converting them back to counts. 
    For the same reason the calibration must be applied uniformely (ie, in the same rest frame) to spectra in different exposures, and their master, so that it does not affect their combinations - even if the original calibration to flux units is applied in the input rest frame.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    print('   > Estimating instrumental calibration')
    data_inst = data_dic[inst] 
        
    #Calculating data
    if gen_dic['calc_gcal']:
        print('         Calculating data')        
        cal_inputs_dic = {} 
        minmax_def = {}
        min_edge_ord_all = np.repeat(1e100,data_dic[inst]['nord'])
        max_edge_ord_all = np.repeat(-1e100,data_dic[inst]['nord'])   
        iexp_glob_groups_vis = {}
        for vis in data_dic[inst]['visit_list']:
            print('           Processing '+vis) 
            data_vis=data_dic[inst][vis]
            data_com_vis = dataload_npz(data_vis['proc_com_data_paths'])
            data_gain_all={}

            #Estimate of instrumental calibration
            #    - set to the requested scaling if error tables are not available as input, or derived from sum(s[F]^2)/sum(F^2) summed over larger bins
            data_vis['cal_data_paths'] = gen_dic['save_data_dir']+'Processed_data/Calibration/'+inst+'_'+vis+'_'
            if data_vis['mock'] or (not gen_dic[inst][vis]['flag_err']):
                if data_vis['mock']:cst_gain = 1.
                elif (not gen_dic[inst][vis]['flag_err']):cst_gain = gen_dic['g_err'][inst]
                data_gain={'gdet_inputs' : {iord : {'par':None,'args':{'constant':cst_gain}} for iord in range(data_dic[inst]['nord'])}}
                n_glob_groups = data_vis['n_in_visit']
                iexp_glob_groups_vis[vis] = range(n_glob_groups) 
                for iexp in range(data_vis['n_in_visit']):data_gain_all[iexp] = data_gain                
    
            else:
                
                #Exposure groups
                iexp_gain_groups = list(range(i,min(i+gen_dic['gcal_binN'],data_vis['n_in_visit'])) for i in range(0,data_vis['n_in_visit'],gen_dic['gcal_binN']))
                n_glob_groups = len(iexp_gain_groups)
                iexp_glob_groups_vis[vis] = range(n_glob_groups)  
                gdet_val_all = np.zeros([data_vis['n_in_visit'],data_dic[inst]['nord']],dtype=object)
                minmax_def[vis] = np.zeros([data_vis['n_in_visit'],data_dic[inst]['nord'],2])*np.nan 
                data_all_temp = {}
                for iord in range(data_dic[inst]['nord']):                    
                    for iexp_glob,iexp_in_group in enumerate(iexp_gain_groups):
                   
                        #Process exposures in current group
                        idx_def_group = np.zeros(0,dtype=int)
                        low_wav_group = np.zeros(0,dtype=float)
                        high_wav_group = np.zeros(0,dtype=float)
                        wav_group = np.zeros(0,dtype=float)
                        flux_group = np.zeros(0,dtype=float)
                        var_group = np.zeros(0,dtype=float)
                        for iexp in iexp_in_group:
                            if iexp not in data_all_temp:data_all_temp[iexp] = dataload_npz(data_vis['proc_DI_data_paths']+str(iexp))
                      
                            #Defined bins
                            #    - we further exclude ranges that were found to display abnormal calibration estimates
                            cond_def_ord = data_all_temp[iexp]['cond_def'][iord]
                            if inst=='HARPN':
                                if data_inst['idx_ord_ref'][iord]==51:cond_def_ord[data_all_temp[iexp]['cen_bins'][iord]<5742.] = False
                                elif data_inst['idx_ord_ref'][iord]==64:cond_def_ord[(data_all_temp[iexp]['cen_bins'][iord]>6561.) & (data_all_temp[iexp]['cen_bins'][iord]<6564.)] = False    
                            elif inst=='HARPS': 
                                if data_inst['idx_ord_ref'][iord]==66:cond_def_ord[(data_all_temp[iexp]['cen_bins'][iord]>6558.) & (data_all_temp[iexp]['cen_bins'][iord]<6569.)] = False
                            idx_def_exp_ord = np_where1D(cond_def_ord)
                    
                            #Concatenate tables so that they are binned together
                            if len(idx_def_exp_ord)>0.:
                                idx_def_group = np.append(idx_def_group,idx_def_exp_ord)
                                low_wav_group = np.append(low_wav_group,data_all_temp[iexp]['edge_bins'][iord,0:-1])
                                high_wav_group = np.append(high_wav_group,data_all_temp[iexp]['edge_bins'][iord,1::] )
                                wav_group = np.append(wav_group,data_all_temp[iexp]['cen_bins'][iord])
                                flux_group = np.append(flux_group,data_all_temp[iexp]['flux'][iord])
                                var_group = np.append(var_group,data_all_temp[iexp]['cov'][iord][0])
               
                                #Save range over which profiles are defined
                                minmax_def[vis][iexp,iord,:]= [data_all_temp[iexp]['edge_bins'][iord,idx_def_exp_ord[0]],data_all_temp[iexp]['edge_bins'][iord,idx_def_exp_ord[-1]]]

                        #Initialize binned tables from grouped exposures
                        if np.sum(idx_def_group)>0:
                            bin_bd,raw_loc_dic = sub_def_bins(gen_dic['gcal_binw'],idx_def_group,low_wav_group,high_wav_group,high_wav_group-low_wav_group,wav_group,flux_group,var1D_loc=var_group)

                            #Adding progressively bins that will be used to fit the correction
                            bin_ord_dic={}
                            for key in ['gdet','cen_bins']:bin_ord_dic[key] = np.zeros(0,dtype=float) 
                            for ibin,(low_bin_loc,high_bin_loc) in enumerate(zip(bin_bd[0:-1],bin_bd[1:])):
                                bin_loc_dic,_ = sub_calc_bins(low_bin_loc,high_bin_loc,raw_loc_dic,0,calc_gdet=True)
                                if len(bin_loc_dic)>0:
                                    bin_ord_dic['gdet'] = np.append( bin_ord_dic['gdet'] , bin_loc_dic['gdet']*1e-3)
                                    bin_ord_dic['cen_bins'] = np.append( bin_ord_dic['cen_bins'] , bin_loc_dic['cen_bins'])
                            gdet_val_all[iexp_glob,iord] = bin_ord_dic
                        else:gdet_val_all[iexp_glob,iord] = None

                    ### End of exposure groups

                ### End of orders 

                #Initialize fit structure
                p_start = Parameters()           
                p_start.add_many(
                          ('a1',0., True   , None , None , None),
                          ('a2',0., True   , None , None , None),
                          ('a3',0., True   , None , None , None),
                          ('a4',0., True   , None , None , None),
                          ('b3',0., True   , None , None , None),
                          ('b4',0., True   , None , None , None),
                          ('c0',0., True   , None , None , None),
                          ('c1',0., True   , None , None , None),
                          ('c2',0., True   , None , None , None),
                          ('c3',0., True   , None , None , None),
                          ('c4',0., True   , None , None , None)) 
                for key in ['blue','mid','red']:
                    if (gen_dic['gcal_deg'][key]<2) or (gen_dic['gcal_deg'][key]>4):stop('Degrees must be between 2 and 4')
                for ideg in range(gen_dic['gcal_deg']['mid']+1,5):p_start['b'+str(ideg)].vary = False                             
                for ideg in range(gen_dic['gcal_deg']['blue']+1,5):p_start['a'+str(ideg)].vary = False                                   
                for ideg in range(gen_dic['gcal_deg']['red']+1,5):p_start['c'+str(ideg)].vary = False   
                nfree_gainfit =  gen_dic['gcal_deg']['blue']+gen_dic['gcal_deg']['red']+1+gen_dic['gcal_deg']['mid']-2                                     

                fixed_args={
                    'use_cov':False,
                    'deg_low':gen_dic['gcal_deg']['blue'],
                    'deg_mid':gen_dic['gcal_deg']['mid'],
                    'deg_high':gen_dic['gcal_deg']['red'],
                    'constant':None                                                
                    }                                           

                #Calibration for spectral profiles
                common_args = (minmax_def[vis],plot_dic,data_dic[inst]['nord'],gdet_val_all,inst,gen_dic['gcal_thresh'][inst],gen_dic['gcal_edges'],gen_dic['gcal_nooutedge'],fixed_args,nfree_gainfit,p_start,data_vis['cal_data_paths'])
                if gen_dic['gcal_nthreads']>1:data_gain_all = para_model_gain(model_gain,gen_dic['gcal_nthreads'],n_glob_groups,[iexp_glob_groups_vis[vis],iexp_gain_groups],common_args)                           
                else:data_gain_all = model_gain(iexp_glob_groups_vis[vis],iexp_gain_groups,*common_args)  

            #Processing all orders for the visit
            cal_inputs_dic[vis] = np.zeros([data_dic[inst]['nord'],n_glob_groups],dtype=object)
            for iord in range(data_dic[inst]['nord']): 

                #Widest spectral range over all visits   
                #    - defined in the input rest frame
                min_edge_ord_all[iord] = np.min([min_edge_ord_all[iord],data_com_vis['min_edge_ord'][iord]])
                max_edge_ord_all[iord] = np.max([max_edge_ord_all[iord],data_com_vis['max_edge_ord'][iord]])
                
                #Retrieve function inputs 
                #    - defined in the input rest frame
                for iexp_glob in iexp_glob_groups_vis[vis]:
                    cal_inputs_dic[vis][iord,iexp_glob] = data_gain_all[iexp_glob]['gdet_inputs'][iord] 

        #Median calibration profile over all exposures in the visit
        #    - we assume the median calibration is smooth enough that it can be captured with an interpolation function
        #    - we assume the spectral tables are always more resolved than the typical variations of the calibration profiles
        #    - the calibration profile is defined in the input rest frame, over a regular table in log oversampled compared to the instrument resolution (dlnw_inst = dw/w = 1/R)
        mean_gdet_func = {} 
        gain_grid_dlnw = 0.5/return_resolv(inst)
        for iord in range(data_dic[inst]['nord']): 

            #Table of definition of mean calibration profile over each order
            #    - over the widest range covered by the inst visits, and at the oversampled instrumental resolution
            #    - tables are uniformely spaced in ln(w)
            #      d[ln(w)] = sc*dw/w = sc*dv/c = sc/R             
            nspec_ord = 1+int( np.ceil(   np.log(max_edge_ord_all[iord]/min_edge_ord_all[iord])/np.log( gain_grid_dlnw + 1. ) )  ) 
            cen_bins_ord = min_edge_ord_all[iord]*( gain_grid_dlnw + 1. )**np.arange(nspec_ord)     
        
            #Median calibration over all exposures in the visit
            #    - interp1d is more stable at the edges than CubicSpline, and capture better the calibration variations than polynomials
            #    - the small variations in measured calibration between orders may result in two slices having different profiles
            #      however this is not an issue for the eventual weighing of the flux profiles
            med_gdet_allvis = np.zeros(nspec_ord,dtype=float)  
            for ivis,vis in enumerate(data_dic[inst]['visit_list']): 
                mean_gdet_ord = np.zeros([nspec_ord,0],dtype=float)*np.nan 
                for iexp_glob in iexp_glob_groups_vis[vis]:
                    mean_gdet_ord_loc = np.zeros(nspec_ord,dtype=float)*np.nan
                    mean_gdet_ord_loc=cal_piecewise_func(cal_inputs_dic[vis][iord,iexp_glob]['par'],cen_bins_ord,args=cal_inputs_dic[vis][iord,iexp_glob]['args'])      
                    mean_gdet_ord = np.append(mean_gdet_ord,mean_gdet_ord_loc[:,None],axis=1)
                med_gdet_allvis+=np.nanmedian(mean_gdet_ord,axis=1)     

            #Mean over all visits
            #    - we limit the calibration below the chosen global outlier threshold
            #    - we do not extrapolate beyond the range of definition of the median calibration profile to avoid spurious behaviour
            med_gdet_ord = 1e3*med_gdet_allvis/gen_dic[inst]['n_visits'] 
            med_gdet_ord[med_gdet_ord>gen_dic['gcal_thresh'][inst]['global']] = gen_dic['gcal_thresh'][inst]['global']
            med_gdet_ord[med_gdet_ord<=0.]=np.min(med_gdet_ord[med_gdet_ord>0.])
            mean_gdet_func[iord] = interp1d(cen_bins_ord,med_gdet_ord,bounds_error=False,fill_value=(med_gdet_ord[0],med_gdet_ord[-1]))  

        #Store mean calibration function
        np.savez_compressed(gen_dic['save_data_dir']+'Processed_data/Calibration/'+inst+'_mean_gdet',data = {'func':mean_gdet_func},allow_pickle=True)  

        #Define calibration tables for each exposure
        #    - the profile is the same, but defined over the table of the exposure
        #    - the path is made specific to a visit and a type of profile so that the calibration function can still be called in the multi-visit routines for any type of profile,
        # even after the type of profile has changed in a given visit
        for vis in data_inst['visit_list']: 
            data_vis=data_inst[vis]
            data_vis['mean_gdet_DI_data_paths'] = {}
            for iexp in range(data_vis['n_in_visit']):
                gain_exp = np.zeros(data_vis['dim_exp'],dtype=float)*np.nan
                data_exp = dataload_npz(data_vis['proc_DI_data_paths']+str(iexp))
                for iord in range(data_inst['nord']):
                    gain_exp[iord] = mean_gdet_func[iord](data_exp['cen_bins'][iord])
                data_vis['mean_gdet_DI_data_paths'][iexp] = data_vis['proc_DI_data_paths']+'mean_gdet_'+str(iexp)
                np.savez_compressed(data_vis['mean_gdet_DI_data_paths'][iexp], data = {'mean_gdet':gain_exp},allow_pickle=True) 
           
    else: 
        for vis in data_inst['visit_list']: 
            data_inst[vis]['cal_data_paths'] = gen_dic['save_data_dir']+'Processed_data/Calibration/'+inst+'_'+vis+'_'
            data_inst[vis]['mean_gdet_DI_data_paths'] = {}
            for iexp in range(data_inst[vis]['n_in_visit']):data_inst[vis]['mean_gdet_DI_data_paths'][iexp] = data_inst[vis]['proc_DI_data_paths']+'mean_gdet_'+str(iexp)
            check_data(data_inst[vis]['mean_gdet_DI_data_paths'],vis=vis)                    

    return None





def model_gain(iexp_glob_groups,iexp_gain_groups,minmax_def,plot_dic,nord,gdet_val_all,inst,gcal_thresh,gcal_edges,gcal_nooutedge,fixed_args,nfree_gainfit,p_start,cal_data_paths):
    r"""**Fit function for instrumental calibration.**

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    data_gain_all = {}
    for iexp_glob,iexp_gain_group in zip(iexp_glob_groups,iexp_gain_groups):
        data_gain={'gdet_inputs':{}}
        if (plot_dic['gcal']!='') or (plot_dic['gcal_ord']!=''):
            data_gain['wav_bin_all']=np.zeros(nord,dtype=object)
            data_gain['wav_trans_all']=np.zeros([2,nord],dtype=float)
            data_gain['cond_fit_all']=np.zeros(nord,dtype=object)
            data_gain['gdet_bin_all']=np.zeros(nord,dtype=object)  
        for iord in range(nord):
            
            #Process order if defined
            fit_done = False
            if gdet_val_all[iexp_glob,iord] is not None:
               
                #Fitting positive values below global threshold
                bin_ord_dic = gdet_val_all[iexp_glob,iord] 
                cond_fit = (bin_ord_dic['gdet']>0.) & (bin_ord_dic['gdet']<gcal_thresh['global']/1e3)
    
                #Remove extreme outliers
                med_prop = np.median(bin_ord_dic['gdet'][cond_fit])
                res = bin_ord_dic['gdet'][cond_fit] - med_prop
                cond_fit[cond_fit][np.abs(res) > 10.*stats.median_abs_deviation(res)] = False  
    
                #Generic fit properties
                #    - the fit does not use the covariance matrix, so fitted tables can be limited to non-consecutive pixels and 'idx_fit' covers all fitted pixels 
                low_edge = bin_ord_dic['cen_bins'][0]
                high_edge = bin_ord_dic['cen_bins'][-1]
                drange =  high_edge-low_edge    
                fixed_args.update({'w_lowedge':low_edge+gcal_edges['blue']*drange,
                                   'w_highedge':high_edge-gcal_edges['red']*drange,
                                   'wref':0.5*(low_edge+high_edge)})  
                
                #Fit binned calibration profile and define complete calibration profile
                #    - if enough bins are defined, otherwise a single measured value is used for the order
                if (np.sum(cond_fit)>nfree_gainfit):
    
                    #Imposing that gradient is negative (resp. positive) at the blue (resp. red) edges
                    p_start.add_many(('dPblue_wmin',-1., True   , None , 0. ))                                            
                    p_start.add_many(('wmin',low_edge-fixed_args['wref'], False)) 
                    p_start.add_many(('a1',0, False   , None , None , 'dPblue_wmin-(2*a2*wmin+3*a3*wmin**2.+4*a4*wmin**3.)')) 
                          
                    p_start.add_many(('dPred_wmax',1., True   , 0. , None))                                            
                    p_start.add_many(('wmax',high_edge-fixed_args['wref'], False)) 
                    p_start.add_many(('c1',0, False   , None , None , 'dPred_wmax-(2*c2*wmax+3*c3*wmax**2.+4*c4*wmax**3.)')) 
                                                          
                    #Temporary fit to identify outliers
                    #    - sigma-clipping applied to the inner parts of the order to prevent excluding edges where calibration can vary sharply
                    #    - negative values are removed
                    #    - weights are used to prevent point with large calibrations (associated with low fluxes) biasing the simple polynomial fit
                    cond_check = deepcopy(cond_fit) 
                    if inst in gcal_nooutedge:cond_check = cond_fit & (bin_ord_dic['cen_bins']>=bin_ord_dic['cen_bins'][0]+gcal_nooutedge[inst][0]) & (bin_ord_dic['cen_bins']<=bin_ord_dic['cen_bins'][-1]-gcal_nooutedge[inst][1]) 
                    if (np.sum(cond_check)>nfree_gainfit):
                        var_fit = bin_ord_dic['gdet'][cond_check]
                        fixed_args['idx_fit'] = np.ones(np.sum(cond_check),dtype=bool)
                        _,merit,_ = call_lmfit(p_start,bin_ord_dic['cen_bins'][cond_check],bin_ord_dic['gdet'][cond_check],np.array([var_fit]),cal_piecewise_func,verbose=False,fixed_args=fixed_args)  
                        res_gdet = bin_ord_dic['gdet'][cond_check] - merit['fit']
                        cond_fit[cond_check] = np.abs(res_gdet)<=gcal_thresh['outliers']*np.std(res_gdet)
    
                        #Model fit
                        #    - errors are scaled with the reduced chi2 from the preliminary fit                           
                        if (np.sum(cond_fit)>nfree_gainfit):
                            fit_done = True
                            var_fit = bin_ord_dic['gdet'][cond_fit]*merit['chi2r'] 
                            fixed_args['idx_fit'] = np.ones(np.sum(cond_fit),dtype=bool)
                            _,merit,p_best = call_lmfit(p_start,bin_ord_dic['cen_bins'][cond_fit],bin_ord_dic['gdet'][cond_fit],np.array([var_fit]),cal_piecewise_func,verbose=False,fixed_args=fixed_args)                                              
                            data_gain['gdet_inputs'][iord] = {'par':deepcopy(p_best),'args':deepcopy(fixed_args)}
    
                #Fit could not be performed
                if not fit_done:
                    data_gain['gdet_inputs'][iord]={'par':None,'args':{'constant':np.median(bin_ord_dic['gdet'])}}
    
                #Save
                if (plot_dic['gcal']!='') or (plot_dic['gcal_ord']!=''):
                    data_gain['wav_bin_all'][iord] = bin_ord_dic['cen_bins']
                    data_gain['wav_trans_all'][:,iord] = [fixed_args['w_lowedge'],fixed_args['w_highedge']]
                    data_gain['cond_fit_all'][iord] = cond_fit
                    data_gain['gdet_bin_all'][iord] = bin_ord_dic['gdet']
    
            #Order fully undefined
            #    - no calibration is applied
            else:
                data_gain['gdet_inputs'][iord]={'par':None,'args':{'constant':1.}}
    
        #Save calibration for each original exposure associated with current exposure group
        data_gain_all[iexp_glob]=data_gain 
        if (plot_dic['gcal']!='') or (plot_dic['gcal_ord']!=''):
            for iexp in iexp_gain_group:np.savez_compressed(cal_data_paths+str(iexp),data=data_gain,allow_pickle=True) 
            
    return data_gain_all

def para_model_gain(func_input,nthreads,n_elem,y_inputs,common_args): 
    r"""**Multithreading routine for model_gain().**

    Args:
        func_input (function): multi-threaded function
        nthreads (int): number of threads
        n_elem (int): number of elements to thread
        y_inputs (list): threadable function inputs 
        common_args (tuple): common function inputs
    
    Returns:
        y_output (None or specific to func_input): function outputs 
    
    """
    pool_proc = Pool(processes=nthreads)   #cannot be passed through lmfit
    ind_chunk_list=init_parallel_func(nthreads,n_elem)
    chunked_args=[(y_inputs[0][ind_chunk[0]:ind_chunk[1]],y_inputs[1][ind_chunk[0]:ind_chunk[1]])+common_args for ind_chunk in ind_chunk_list]	
    all_results=tuple(tab for tab in pool_proc.starmap(func_input,chunked_args))

    data_gain_all = {}
    for data_gain in tuple(all_results[i] for i in range(nthreads)):data_gain_all.update(data_gain)
    y_output=data_gain_all
    
    pool_proc.close()
    pool_proc.join() 				
    return y_output




def cal_piecewise_func(param_in,wav_in,args=None):
    r"""**Instrumental calibration model.**
    
    Defines joined polynomials to model calibration profiles.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    if isinstance(param_in,lmfit.parameter.Parameters):params={par:param_in[par].value for par in param_in}
    else:params=deepcopy(param_in)                                                   
    #P_low(w) = sum(0:nlow,ai*w^i)
    #P_mid(w) = sum(0:nmid,bi*w^i)                                               
    #P_high(w) = sum(0:nhigh,ci*w^i)
    #   with w = wav-wav_ref
    #
    #Continuity in derivative:
    #P_low'(w1) = P_mid'(w1)  
    #   sum(1:nlow,i*ai*w1^i-1) = b1 + 2*b2*w1 + sum(3:nmid,i*bi*w1^i-1)
    #P_high'(w2) = P_mid'(w2)   
    #   sum(1:nhigh,i*ci*w2^i-1) = b1 + 2*b2*w2 + sum(3:nmid,i*bi*w2^i-1)
    #
    # 2*b2*(w1-w2) = sum(1:nlow,i*ai*w1^i-1) - sum(1:nhigh,i*ci*w2^i-1) + sum(3:nmid,i*bi*w2^i-1) - sum(3:nmid,i*bi*w1^i-1)
    # > b2 = (sum(1:nlow,i*ai*w1^i-1) - sum(1:nhigh,i*ci*w2^i-1) + sum(3:nmid,i*bi*w2^i-1) - sum(3:nmid,i*bi*w1^i-1))/(2*(w1-w2))      
    #      = (sum(0:nlow-1,(j+1)*a[j+1]*w1^j) - sum(0:nhigh-1,(j+1)*c[j+1]*w2^j) + sum(2:nmid-1,(j+1)*b[j+1]*w2^j) - sum(2:nmid-1,(j+1)*b[j+1]*w1^j)  )/(2*(w1-w2))     
    #
    # > b1 = sum(0:nlow-1,(j+1)*a[j+1]*w1^j) - sum(2:nmid,i*bi*w1^i-1)
    #      = sum(0:nlow-1,(j+1)*a[j+1]*w1^j) - sum(1:nmid-1,(j+1)*b[j+1]*w1^j) 
    #
    #Continuity in value:
    #P_low(w1) = P_mid(w1) 
    #   sum(0:nlow,ai*w1^i) = b0 + sum(1:nmid,bi*w1^i)
    #P_high(w2) = P_mid(w2)   
    #   sum(0:nhigh,ci*w2^i) = b0 + sum(1:nmid,bi*w2^i)
    #
    # a0 + sum(1:nlow,ai*w1^i) - sum(0:nhigh,ci*w2^i) = sum(1:nmid,bi*w1^i) - sum(1:nmid,bi*w2^i)  
    # > a0 = sum(1:nmid,bi*w1^i) - sum(1:nmid,bi*w2^i) - sum(1:nlow,ai*w1^i) + sum(0:nhigh,ci*w2^i)
    #
    # > b0 = sum(0:nlow,ai*w1^i) - sum(1:nmid,bi*w1^i)
    if args['constant'] is not None:
        model = np.repeat(args['constant'],len(wav_in))
    else:
        wav = wav_in-args['wref']
        w1 = args['w_lowedge']-args['wref']
        w2 = args['w_highedge']-args['wref'] 
    
        #Derivative continuity
        dPlow = Polynomial([(ideg+1)*params['a'+str(ideg+1)] for ideg in range(args['deg_low'])])
        dPhigh = Polynomial([(ideg+1)*params['c'+str(ideg+1)] for ideg in range(args['deg_high'])])
        dPmid_cut = Polynomial([0.,0.]+[(ideg+1)*params['b'+str(ideg+1)] for ideg in range(2,args['deg_mid'])])
        params['b2'] = (dPlow(w1) - dPhigh(w2) + dPmid_cut(w2) - dPmid_cut(w1))/(2.*(w1-w2)) 
        params['b1'] = dPlow(w1) - Polynomial([0.]+[(ideg+1)*params['b'+str(ideg+1)] for ideg in range(1,args['deg_mid'])])(w1)   

        #Value continuity        
        Phigh = Polynomial([params['c'+str(ideg)] for ideg in range(args['deg_high']+1)])   
        Pmid_cut = Polynomial([0.]+[params['b'+str(ideg)] for ideg in range(1,args['deg_mid']+1)])
        params['a0'] = Pmid_cut(w1) - Pmid_cut(w2) - Polynomial([0]+[params['a'+str(ideg)] for ideg in range(1,args['deg_low']+1)])(w1) + Phigh(w2)
        Plow = Polynomial([params['a'+str(ideg)] for ideg in range(args['deg_low']+1)])
        params['b0'] = Plow(w1) - Pmid_cut(w1)
    
        #Model
        model = np.zeros(len(wav),dtype=float)
        cond_mid = (wav>=w1) & (wav<=w2)
        model[cond_mid] = Polynomial([params['b'+str(ideg)] for ideg in range(args['deg_mid']+1)])(wav[cond_mid])                          
        cond_lowe=wav<w1
        model[cond_lowe] = Plow(wav[cond_lowe])     
        cond_highe=wav>w2
        model[cond_highe] = Phigh(wav[cond_highe])
        
    return model

